/* backend.h - header for the jvterm display engine
 *
 * Copyright (C) 2003, 2005  Jochen Voss
 *
 * This program is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 2 of the License, or
 * (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 51 Franklin St, Fifth Floor, Boston, MA  02110-1301 USA
 *
 * $Id: backend.h 6725 2005-11-09 23:15:58Z voss $
 */

#ifndef FILE_BACKEND_H_SEEN
#define FILE_BACKEND_H_SEEN

#include <SDL.h>


/* charcell.c */

/* bit masks for character attributes */
#define JV_ATTR_BLINK 	1
#define JV_ATTR_BRIGHT 	2
#define JV_ATTR_DIM 	4
#define JV_ATTR_REVERSE	8
#define JV_ATTR_UNDERLINE 16
#define JV_ATTR_ACS	32
#define JV_ATTR_COMBINED 64
/* 128 is reserved for cursor handling */

struct charcell {
  Uint16 c;			/* the character value encoded in UCS-2 */
  Uint8 attr;			/* the bitset of character attributes */
  Uint8 colors;			/* foreground/background color pair */
};

extern  int  char_is_combining (Uint16 n) jv_const;
extern  struct charcell  charcell_combine (struct charcell cc1, Uint16 c2);
extern  int  charcell_disband (struct charcell cc, Uint16 *c1, Uint16 *c2);
extern  char *charcell_attr_name (Uint8 attr);


/* color.c */

#define JV_COLOR_BLACK  0
#define JV_COLOR_RED    1
#define JV_COLOR_GREEN  2
#define JV_COLOR_YELLOW 3
#define JV_COLOR_BLUE   4
#define JV_COLOR_MAGENTA 5
#define JV_COLOR_CYAN   6
#define JV_COLOR_WHITE  7

#define GET_COLOR_FG(colors) ((colors) & 15)
#define GET_COLOR_BG(colors) ((colors) >> 4)
#define JV_COLOR_PAIR(fg,bg) (16*(bg)+fg)

#define DEFAULT_COLOR JV_COLOR_PAIR (JV_COLOR_WHITE, JV_COLOR_BLACK)

extern  const char *color_name (Uint8 col);
extern  void  color_decode (struct charcell cc,
			    SDL_Color **fg, SDL_Color **bg);


/* font.c */

struct font_info {
  int  cell_width;		/* character cell width (pixels) */
  int  cell_height;		/* character cell height (pixels) */
  int  ascent, descent;		/* pixels above/below baseline */
  double  ula, ulb;		/* upper and lower position of underline */
};

extern  void  load_font (const char *ttf_file_name, double ptsize,
			 struct font_info *font_info);
extern  void  unload_font (void);

extern  SDL_Surface *font_get_glyph (struct font_info *fi,
				     const struct charcell cc);

/* screen.c */

extern  const struct charcell  default_char;

extern  void  open_screen (int depth, int hw_flag, int fullscreen_flag,
			   const char *font_file, double font_size);
extern  void  close_screen (void);
extern  int  screen_width (void);
extern  int  screen_height (void);
extern  int  screen_get_cols (void);
extern  int  screen_get_rows (void);

extern  void  screen_update (void);
extern  void  screen_set_cursor (int x, int y);
extern  void  screen_toggle_cursor (void);
extern  void  screen_toggle_blink (void);
extern  void  screen_toggle_flash (void);

extern  void  screen_clear (void);
extern  struct charcell  screen_get_cell (int x, int y);
extern  void  screen_draw_cell (struct charcell cc, int x, int y);
extern  void  screen_scroll_up (int y0, int y1);
extern  void  screen_scroll_down (int y0, int y1);

extern  void  screen_set_selection (int point_x, int point_y,
				    int mark_x, int mark_y);

extern  void  screen_set_image (SDL_Surface *image, int xoffs, int yoffs);

/* image.c */

struct image {
  SDL_Surface *orig;
  SDL_Surface *scaled;
  int  xoffs, yoffs;
};

extern  struct image *load_image (const char *file_name);
extern  void  delete_image (struct image *im);

extern  SDL_Surface *image_get (struct image *im);
extern  void  image_scale (struct image *im, double scale);
extern  void  image_scroll (struct image *im, int dx, int dy);

/* buffer.c */

struct buffer {
  int  active_flag;		/* is the buffer connected to a screen? */

  struct buffer_row *lines;	/* array of lines in the buffer */
  int  cols;			/* width of buffer */
  int  screen_rows;		/* number of visible lines */
  int  total_rows;		/* allocated number of buffer rows */

  int  used_rows;		/* number of used lines */
  int  pos;			/* index of first visible line */

  int  cursor_flag;		/* is the cursor enabled? */
  int  cx, cy;			/* cursor position in screen coordinates */

  int  am_flag;			/* are automatic margins enabled? */

  int  sel_point;		/* position of the selection point */
  int  sel_mark;		/* position of the selection mark */
  int  basechar;		/* basechar for combining characters */

  struct image *im;
};

extern  struct buffer *create_buffer (int scrollback);
extern  void  delete_buffer (struct buffer *b);

extern  void  buffer_connect (struct buffer *b);
extern  void  buffer_disconnect (struct buffer *b);
extern  void  buffer_reset (struct buffer *b);
extern  void  buffer_clear (struct buffer *b);

extern  void  buffer_lf (struct buffer *b);
extern  void  buffer_crlf (struct buffer *b);
extern  void  buffer_putc (struct buffer *b, struct charcell cc);
extern  void  buffer_puts (struct buffer *b, const char *str,
			   Uint8 attr, Uint8 colors);
extern  void  buffer_clear_basechar (struct buffer *b);

extern  void  buffer_scroll_up (struct buffer *b);
extern  void  buffer_scroll_down (struct buffer *b);
extern  void  buffer_insert_line (struct buffer *b);
extern  void  buffer_delete_line (struct buffer *b);

extern  void  buffer_clear_rect (struct buffer *b,
				 int x0, int y0, int x1, int y1);
extern  void  buffer_delete_char (struct buffer *b, int x, int y, int n);
extern  void  buffer_insert_char (struct buffer *b, int x, int y, int n);

extern  void  buffer_toggle_cursor (struct buffer *b, int flag);
extern  void  buffer_set_cursor (struct buffer *b, int x, int y);

extern  void  buffer_vscroll_back (struct buffer *b, int n);
extern  void  buffer_vscroll_forward (struct buffer *b, int n);
extern  void  buffer_vscroll_reset (struct buffer *b);

extern  void  buffer_selection_start (struct buffer *b);
extern  void  buffer_selection_stop (struct buffer *b);
extern  Uint16 *buffer_selection_get (struct buffer *b);
extern  void  buffer_mark_set (struct buffer *b);
extern  void  buffer_point_up (struct buffer *b);
extern  void  buffer_point_down (struct buffer *b);
extern  void  buffer_point_left (struct buffer *b);
extern  void  buffer_point_right (struct buffer *b);

extern  void  buffer_image_load (struct buffer *b, const char *file_name);
extern  void  buffer_image_zoom (struct buffer *b, double q);
extern  void  buffer_image_scroll (struct buffer *b, int dx, int dy);

#endif /* FILE_BACKEND_H_SEEN */
