/* jvterm.h - header file for the jvterm terminal emulator's base layer
 *
 * Copyright (C) 2003, 2005  Jochen Voss
 *
 * This program is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 2 of the License, or
 * (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 51 Franklin St, Fifth Floor, Boston, MA  02110-1301 USA
 *
 * $Id: base.h 6725 2005-11-09 23:15:58Z voss $
 */

#ifndef FILE_JVTERM_H_SEEN
#define FILE_JVTERM_H_SEEN

#include <stdlib.h>
#include <unistd.h>

#if __GNUC__ >= 3
#define  inline  inline __attribute__ ((always_inline))
#define  jv_pure  __attribute__ ((pure))
#define  jv_const  __attribute__ ((const))
#define  jv_noreturn  __attribute__ ((noreturn))
#define  jv_malloc  __attribute__ ((malloc))
#define  jv_printf  __attribute__ ((format (printf, 1, 2)))
#define  jv_must_check  __attribute__ ((warn_unused_result))
#define  jv_deprecated  __attribute__ ((deprecated))
#define  jv_unused  __attribute__ ((unused))
#else
#define  inline  /* no inline */
#define  jv_pure  /* no pure */
#define  jv_const  /* no const */
#define  jv_noreturn  /* no noreturn */
#define  jv_malloc  /* no malloc */
#define  jv_printf  /* no printf */
#define  jv_must_check  /* no warn_unused_result */
#define  jv_deprecated  /* no deprecated */
#define  jv_unused  /* no unused */
#endif

#define min(a,b) \
       ({ typeof (a) _a = (a); \
          typeof (b) _b = (b); \
          _a < _b ? _a : _b; })

#define max(a,b) \
       ({ typeof (a) _a = (a); \
          typeof (b) _b = (b); \
          _a > _b ? _a : _b; })

/* xmalloc.c */

extern  void *xmalloc (size_t size);
#define xnew(T) ((T *)xmalloc(sizeof(T)))
extern  void *xrealloc (void *ptr, size_t newsize);
extern  void  xfree (void *ptr);
extern  char *xstrdup (const char *s);

/* log.c */

/* log levels:
 *   log_MESSAGE = debugging and status information,
 *                 not normally presented to the user
 *                 (may be emitted via stdout)
 *   log_WARNING = warnings, only presented to the user in
 *                 verbose mode (may be emitted via stderr)
 *   log_ERROR   = errors which the user should know about
 *                 (may be emitted via stderr)
 */
enum loglevel { log_MESSAGE, log_WARNING, log_ERROR };
typedef  void  (*write_line_fn) (enum loglevel level, const char *message,
				 void *client_data);

extern  void  open_log (void);
extern  void  close_log (void);

extern  void  log_up_and_running (void);
extern  int  log_add_client (write_line_fn write_line, void *client_data);
extern  void  log_remove_client (int handle);

extern  void  log_write_line (enum loglevel level, const char *message);

/* error.c */
extern  void  fatal (const char *format, ...) jv_noreturn jv_printf;
extern  void  fatal_sdl (const char *str) jv_noreturn;
extern  void  error (const char *format, ...) jv_printf;
extern  void  error_sdl (const char *str);
extern  void  warning (const char *format, ...) jv_printf;
extern  void  message (const char *format, ...) jv_printf;

/* options.c */

struct voption {
  const char  *name;		/* long option name */
  char  short_name;		/* short option name */
  int *flag_ptr;		/* where is the flag */
  int  has_arg;			/* Is there an argument? */
  const char *arg_name;		/* name of the argument */
  const char *help;		/* "--help" line */
};

#define V_MIXED		1
#define V_ONE_HYPHEN	2
#define V_KEEP_UNKNOWN	4
#define V_NO_JOIN	8

extern  void  open_options (int argc, char **argv);
extern  void  close_options (void);
extern  int  options_get (const struct voption *options,
			  const char **argptr, int flags);
extern  void  options_show (const struct voption *options);

/* converter.c */
struct converter;
typedef  ssize_t (*write_fn) (const void *buffer, size_t size,
			      void *client_data);

extern struct converter *open_converter(const char * id,
					const char *from, const char *to,
					const void *repl, size_t repl_len,
					write_fn write, void *client_data);
extern  void  close_converter (struct converter *conv);

extern  void  converter_reset (struct converter *conv);
extern  void  converter_write (struct converter *conv,
			       const void *buffer, size_t size);
extern  ssize_t  converter_copy (struct converter *conv, int fd);

#endif /* FILE_JVTERM_H_SEEN */
