/* buffer.c - implement a framebuffer for character data
 *
 * Copyright (C) 2003, 2005  Jochen Voss
 *
 * This program is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 2 of the License, or
 * (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 51 Franklin St, Fifth Floor, Boston, MA  02110-1301 USA
 *
 * $Id: buffer.c 6725 2005-11-09 23:15:58Z voss $
 */

#ifdef HAVE_CONFIG_H
#  include <config.h>
#endif

#include <stdlib.h>
#include <string.h>

#include "base.h"
#include "backend.h"


struct buffer_row {
  struct charcell *data;
};


static void
buffer_draw_cursor (const struct buffer *b)
/* Update the cursor position on the screen.  */
{
  if (! b->active_flag)  return;

  if (b->cursor_flag) {
    int  real_y = b->cy + b->used_rows - b->screen_rows - b->pos;
    screen_set_cursor (b->cx, real_y);
  } else {
    screen_set_cursor (0, -1);
  }
}

static struct charcell
buffer_get (const struct buffer *b, int x, int y)
/* Get the character from position (x,y) in screen coordinates.  */
{
  return  b->lines[y+b->pos].data[x];
}

static void
redraw_cell (const struct buffer *b, int x, int y)
/* Update a character cell on the screen, potentially while virtual
 * scrolling is in effect.  */
{
  struct charcell  cc;

  if (! b->active_flag)  return;

  cc = buffer_get (b, x, y);
  screen_draw_cell (cc, x, y);
}

static void
buffer_draw_selection (const struct buffer *b)
{
  int  point = b->sel_point;
  int  mark = b->sel_mark;
  int  min = b->pos*b->cols;
  int  max = min + b->screen_rows*b->cols;

  if (! b->active_flag)  return;

  if (mark < 0)  mark = point;
  if (point < 0
      || (point<min && mark<min)
      || (point>max && mark>max)) {
    screen_set_selection (0, -1, 0, 0);
  } else {
    if (point > max)  point = max;
    if (mark > max)  mark = max;
    point -= min;
    mark -= min;
    screen_set_selection (point % b->cols, point / b->cols,
			  mark % b->cols, mark / b->cols);
  }
}

static void
buffer_draw (const struct buffer *b)
{
  int  x, y;
  
  for (y=0; y<b->screen_rows; ++y) {
    for (x=0; x<b->cols; ++x) {
      redraw_cell (b, x, y);
    }
  }
  buffer_draw_cursor (b);
  buffer_draw_selection (b);
}

static void
buffer_set (struct buffer *b, struct charcell cc, int x, int y)
/* Place a character on the screen.
 * This may not be called while virtual scrolling is in effect.  */
{
  b->lines[y+b->pos].data[x] = cc;
  redraw_cell (b, x, y);
}

/**********************************************************************
 * user visible functions
 */

struct buffer *
create_buffer (int scrollback)
/* Create a new buffer with SCROLLBACK lines of scroll back buffer.
 * The newly created buffer is not yet connected to the screen.  */
{
  struct buffer *b;
  int  i;

  b = xnew (struct buffer);

  b->active_flag = 0;
  
  b->cols = screen_get_cols ();
  b->screen_rows = screen_get_rows ();
  b->total_rows = b->screen_rows + scrollback;
  b->lines = xmalloc (sizeof(struct buffer_row) * b->total_rows);
  for (i=0; i<b->total_rows; ++i) {
    b->lines[i].data = xmalloc (b->cols*sizeof (struct charcell));
  }

  b->im = NULL;

  buffer_reset (b);
  
  return b;
}

void
delete_buffer (struct buffer *b)
{
  int  i;

  if (b->im)  delete_image (b->im);
  
  for (i=0; i<b->total_rows; ++i)  xfree (b->lines[i].data);
  xfree (b->lines);
  xfree (b);
}

void
buffer_connect (struct buffer *b)
/* Connect the buffer B to the screen.
 * The caller must make sure to connected only one buffer at a time.  */
{
  b->active_flag = 1;
  if (b->im)  screen_set_image (image_get(b->im), b->im->xoffs, b->im->yoffs);
  buffer_draw (b);
}

void
buffer_disconnect (struct buffer *b)
/* Disconnect a buffer from the screen.  */
{
  if (b->im)  screen_set_image (NULL, 0, 0);
  b->active_flag = 0;
}

void
buffer_reset (struct buffer *b)
/* Reset the buffer to its original state.  */
{
  b->cursor_flag = 1;
  b->am_flag = 1;

  buffer_image_load (b, NULL);
  buffer_clear (b);
}

void
buffer_clear (struct buffer *b)
/* Clear the buffer and return the cursor to the home position.  */
{
  int  i, j;

  b->pos = 0;
  b->used_rows = b->screen_rows;
  
  for (j=0; j<b->used_rows; ++j) {
    for (i=0; i<b->cols; ++i) {
      b->lines[j].data[i] = default_char;
    }
  }
  if (b->active_flag)  screen_clear ();

  b->sel_point = -1;
  b->sel_mark = -1;
  b->basechar = -1;

  buffer_set_cursor (b, 0, 0);
}

void
buffer_lf (struct buffer *b)
/* Move the cursor one row down.  Scroll the screen as needed.  */
{
  int  x, y;
  
  x = b->cx;
  y = b->cy;
  if (y < b->screen_rows - 1) {
    buffer_set_cursor (b, x, y+1);
  } else {
    buffer_scroll_up (b);
    buffer_set_cursor (b, x, y);
  }
}

void
buffer_crlf (struct buffer *b)
/* Move the cursor to the beginning of the next row.  Scroll the
 * screen as needed  */
{
  int  y;
  
  y = b->cy;
  if (y < b->screen_rows - 1) {
    buffer_set_cursor (b, 0, y+1);
  } else {
    buffer_scroll_up (b);
    buffer_set_cursor (b, 0, y);
  }
}

void
buffer_putc (struct buffer *b, struct charcell cc)
/* Insert CC into the buffer at the current cursor position.
 * Move the cursor to the next cursor position.  This obeys
 * the am_flag setting and scrolls the screen as needed.  */
{
  buffer_vscroll_reset (b);
  if (b->basechar >= 0 && char_is_combining (cc.c)) {
    struct charcell  bc;
    int  x, y;
    x = b->basechar % b->cols;
    y = b->basechar / b->cols - b->pos;
    bc = buffer_get (b, x, y);
    bc = charcell_combine (bc, cc.c);
    buffer_set (b, bc, x, y);
  } else {
    buffer_set (b, cc, b->cx, b->cy);
    b->basechar = b->cx + b->cols*(b->cy + b->pos);
    if (b->cx < b->cols-1) {
      buffer_set_cursor (b, b->cx+1, b->cy);
    } else if (b->am_flag) {
      buffer_crlf (b);
    }
  }
}

void
buffer_puts (struct buffer *b, const char *str, Uint8 attr, Uint8 colors)
{
  struct charcell  cc;

  cc.attr = attr;
  cc.colors = colors;

  while (*str) {
    if (*str == '\n') {
      buffer_crlf (b);
    } else {
      cc.c = *str;
      buffer_putc (b, cc);
    }
    ++str;
  }
}

void
buffer_clear_basechar (struct buffer *b)
{
  b->basechar = -1;
}

void
buffer_scroll_up (struct buffer *b)
/* Scroll up the buffer contents by one line.
 * This does not affect the cursor.  */
{
  int  i, j;
  
  buffer_vscroll_reset (b);

  if (b->used_rows < b->total_rows) {
    b->used_rows += 1;
    b->pos += 1;
  } else {
    struct buffer_row tmp;
    
    tmp = b->lines[0];
    j = b->used_rows-1;
    memmove (b->lines, b->lines+1, j*sizeof(struct buffer_row));
    b->lines[j] = tmp;
  }
  j = b->used_rows-1;
  for (i=0; i<b->cols; ++i)  b->lines[j].data[i] = default_char;

  if (b->active_flag)  screen_scroll_up (0, b->screen_rows);
  j = b->screen_rows-1;
  for (i=0; i<b->cols; ++i)  redraw_cell (b, i, j);
}

void
buffer_scroll_down (struct buffer *b)
/* Scroll down the buffer contents by one line.
 * This does not affect the cursor. */
{
  int  i, j;
  
  buffer_vscroll_reset (b);

  if (b->pos > 0) {
    b->used_rows -= 1;
    b->pos -= 1;
  } else {
    struct buffer_row tmp;
    
    j = b->used_rows-1;
    tmp = b->lines[j];
    memmove (b->lines+1, b->lines, j*sizeof(struct buffer_row));
    b->lines[0] = tmp;
    for (i=0; i<b->cols; ++i)  tmp.data[i] = default_char;
  }

  if (b->active_flag) {
    screen_scroll_down (0, b->screen_rows);
    for (i=0; i<b->cols; ++i)  redraw_cell (b, i, 0);
  }
}

void
buffer_insert_line (struct buffer *b)
/* Insert a new empty line at the cursor position.
 * The lower part of the screen is scolled down, the last line is lost.
 * This does not affect the cursor position.  */
{
  int  i, k0, k1;
  struct buffer_row tmp;

  buffer_vscroll_reset (b);

  k0 = b->pos + b->cy;
  k1 = b->used_rows;
  tmp = b->lines[k1-1];
  memmove (b->lines+k0+1, b->lines+k0, (k1-k0-1)*sizeof(struct buffer_row));
  b->lines[k0] = tmp;
  for (i=0; i<b->cols; ++i)  tmp.data[i] = default_char;

  if (b->active_flag)  screen_scroll_down (b->cy, b->screen_rows);
  for (i=0; i<b->cols; ++i)  redraw_cell (b, i, b->cy);
}

void
buffer_delete_line (struct buffer *b)
/* Delete the line at the cursor position.
 * The lower part of the screen is scolled up.
 * This does not affect the cursor position.  */
{
  int  i, k0, k1;
  struct buffer_row tmp;

  buffer_vscroll_reset (b);

  k0 = b->pos + b->cy;
  k1 = b->used_rows;
  tmp = b->lines[k0];
  memmove (b->lines+k0, b->lines+k0+1, (k1-k0-1)*sizeof(struct buffer_row));
  b->lines[k1-1] = tmp;
  for (i=0; i<b->cols; ++i)  tmp.data[i] = default_char;

  if (b->active_flag)  screen_scroll_up (b->cy, b->screen_rows);
  for (i=0; i<b->cols; ++i)  redraw_cell (b, i, b->screen_rows-1);
}

void
buffer_clear_rect (struct buffer *b, int x0, int y0, int x1, int y1)
/* Clear a rectangular region of the screen.
 * This does not affect cursor.  */
{
  int  x, y;

  buffer_vscroll_reset (b);
  
  for (y=y0; y<y1; ++y) {
    for (x=x0; x<x1; ++x) {
      buffer_set (b, default_char, x, y);
    }
  }
}

void
buffer_delete_char (struct buffer *b, int x, int y, int n)
/* Delete the character under the cursor and move the
 * righthand part of the line one position to the left.  This
 * does not affect cursor.  */
{
  int  i;

  buffer_vscroll_reset (b);

  for (i=x; i<b->cols; ++i) {
    if (i+n < b->cols) {
      struct charcell cc;
      cc = buffer_get (b, i+n, y);
      buffer_set (b, cc, i, y);
    } else {
      buffer_set (b, default_char, i, y);
    }
  }
}

void
buffer_insert_char (struct buffer *b, int x, int y, int n)
/* Insert a space at the cursor position and move the
 * righthand part of the line one position to the right.  The
 * rightmost character of the line is lost.  This does not
 * affect the cursor position.  */
{
  int  i;

  buffer_vscroll_reset (b);

  for (i=b->cols-1; i>=x; --i) {
    if (i < x+n) {
      buffer_set (b, default_char, i, y);
    } else {
      struct charcell cc;
      cc = buffer_get (b, i-n, y);
      buffer_set (b, cc, i, y);
    }
  }
}

void
buffer_toggle_cursor (struct buffer *b, int flag)
{
  b->cursor_flag = flag;
  buffer_draw_cursor (b);
}

void
buffer_set_cursor (struct buffer *b, int x, int y)
{
  b->cx = x;
  b->cy = y;
  buffer_draw_cursor (b);
}

void
buffer_vscroll_back (struct buffer *b, int n)
/* Use virtual scrolling to scroll the buffer backward by N lines.
 * The special value N=0 means half of a screen heigth.  */
{
  if (n <= 0)  n = b->screen_rows/2;
  if (n > b->pos)  n = b->pos;
  b->pos -= n;

  buffer_draw (b);
}

void
buffer_vscroll_forward (struct buffer *b, int n)
/* Use virtual scrolling to scroll forward the buffer by N lines.
 * The value N=0 means half the screen height.  */
{
  int  limit;

  limit = b->used_rows - b->screen_rows;

  if (n <= 0)  n = b->screen_rows/2;
  if (b->pos+n > limit)  n = limit - b->pos;
  b->pos += n;

  buffer_draw (b);
}

void
buffer_vscroll_reset (struct buffer *b)
/* Stop doing virtual scrolling and reset the visual screen
 * position to match the buffer's screen position.  */
{
  if (b->pos != b->used_rows - b->screen_rows) {
    b->pos = b->used_rows - b->screen_rows;
    buffer_draw (b);
  }
}

/**********************************************************************
 * selection handling
 */

void
buffer_selection_start (struct buffer *b)
{
  if (b->cursor_flag
      && b->cy + b->used_rows - b->screen_rows - b->pos >= 0
      && b->cy + b->used_rows - b->screen_rows - b->pos < b->screen_rows) {
    b->sel_point = b->cx + b->cols*(b->cy + b->used_rows - b->screen_rows);
  } else {
    b->sel_point = b->cols*(b->screen_rows+b->pos-1);
  }
  b->sel_mark = -1;
  buffer_draw_selection (b);
}

void
buffer_selection_stop (struct buffer *b)
{
  b->sel_point = -1;
  b->sel_mark = -1;
  b->pos = b->used_rows - b->screen_rows;
  buffer_draw (b);
}

Uint16 *
buffer_selection_get (struct buffer *b)
{
  int  ax, ay, bx, by;
  int  i, j, n;
  Uint16 *res;

  if (b->sel_mark < 0)  return NULL;
  
  if (b->sel_point < b->sel_mark) {
    ax = b->sel_point % b->cols;
    ay = b->sel_point / b->cols;
    bx = b->sel_mark % b->cols;
    by = b->sel_mark / b->cols;
  } else {
    ax = b->sel_mark % b->cols;
    ay = b->sel_mark / b->cols;
    bx = b->sel_point % b->cols;
    by = b->sel_point / b->cols;
  }

  n = 0;
  for (j=ay; j<=by; ++j) {
    int  min;

    min = (j==ay) ? ax : 0;
    if (j == by) {
      i = bx;
    } else {
      for (i=b->cols; i>0 && i>min; --i) {
	if (b->lines[j].data[i-1].c != default_char.c)  break;
      }
    }
    n += i-min;
    if (j<by) ++n;		/* space for the trailing NL */
  }
  if (n == 0)  return NULL;

  ++n;				/* space for the trailing NUL */
  
  res = xmalloc (n*sizeof(Uint16));
  n = 0;
  for (j=ay; j<=by; ++j) {
    int  min, max;

    min = (j==ay) ? ax : 0;
    if (j == by) {
      i = bx;
    } else {
      for (i=b->cols; i>0 && i>min; --i) {
	if (b->lines[j].data[i-1].c != default_char.c)  break;
      }
    }
    max = i;
    for (i=min; i<max; ++i)  res[n++] = b->lines[j].data[i].c;
    if (j<by) res[n++] = '\r';
  }
  res[n++] = '\0';
  
  return  res;
}

void
buffer_mark_set (struct buffer *b)
{
  b->sel_mark = b->sel_point;
}

void
buffer_point_up (struct buffer *b)
{
  if (b->sel_point >= b->cols) {
    b->sel_point -= b->cols;
    if (b->sel_point / b->cols < b->pos) {
      b->pos = b->sel_point / b->cols;
      buffer_draw (b);
    } else {
      buffer_draw_selection (b);
    }
  }
}

void
buffer_point_down (struct buffer *b)
{
  if (b->sel_point <= (b->used_rows-1)*b->cols) {
    b->sel_point += b->cols;
    if (b->sel_point / b->cols >= b->pos + b->screen_rows
	&& b->pos + b->screen_rows < b->used_rows) {
      b->pos = b->sel_point / b->cols - b->screen_rows + 1;
      buffer_draw (b);
    } else {
      buffer_draw_selection (b);
    }
  }
}

void
buffer_point_left (struct buffer *b)
{
  if (b->sel_point > 0) {
    b->sel_point -= 1;
    if (b->sel_point / b->cols < b->pos) {
      b->pos = b->sel_point / b->cols;
      buffer_draw (b);
    } else {
      buffer_draw_selection (b);
    }
  }
}

void
buffer_point_right (struct buffer *b)
{
  if (b->sel_point < b->used_rows*b->cols) {
    b->sel_point += 1;
    if (b->sel_point / b->cols >= b->pos + b->screen_rows
	&& b->pos + b->screen_rows < b->used_rows) {
      b->pos = b->sel_point / b->cols - b->screen_rows + 1;
      buffer_draw (b);
    } else {
      buffer_draw_selection (b);
    }
  }
}

/**********************************************************************
 * image handling
 */

void
buffer_image_load (struct buffer *b, const char *file_name)
{
  if (file_name && *file_name) {
    b->im = load_image (file_name);
    if (b->im && b->active_flag) {
      screen_set_image (image_get(b->im), b->im->xoffs, b->im->yoffs);
    }
  } else if (b->im) {
    if (b->active_flag)  screen_set_image (NULL, 0, 0);
    delete_image (b->im);
    b->im = NULL;
  }
}

void
buffer_image_zoom (struct buffer *b, double q)
{
  if (! b->im)  return;
  image_scale (b->im, q);
  if (b->active_flag) {
    screen_set_image (image_get(b->im), b->im->xoffs, b->im->yoffs);
  }
}

void
buffer_image_scroll (struct buffer *b, int dx, int dy)
{
  if (! b->im)  return;
  image_scroll (b->im, dx, dy);
  if (b->active_flag) {
    screen_set_image (image_get(b->im), b->im->xoffs, b->im->yoffs);
  }
}

/**********************************************************************
 * unit testing
 */

#ifdef TEST_BUFFER

#include <stdio.h>
#include <unistd.h>
#include <jvrand.h>

int
main (int argc, char **argv)
{
  int  error_flag = 0;
  int  hw_flag = 0;
  int  depth = 0;
  int  font_size = 15;
  const char *font_file = NULL;
  int  fullscreen_flag = 1;

  struct buffer *b1, *b2, *b3;
  
  do {
    int  c;
    c = getopt (argc, argv, "db:f:p:qw");
    if (c == -1)  break;

    switch (c) {
    case 'b':
      depth = strtol (optarg, NULL, 0);
      break;
    case 'f':
      font_file = xstrdup (optarg);
      break;
    case 'q':
      hw_flag = 1;
      break;
    case 'p':
      font_size = strtol (optarg, NULL, 0);
      break;
    case 'w':
      fullscreen_flag = 0;
      break;
    default:
      error_flag = 1;
    }
  } while (! error_flag);
  
  if ((SDL_Init (SDL_INIT_VIDEO|SDL_INIT_TIMER) == -1)) {
    fatal_sdl ("Could not initialize SDL");
  }
  
  open_screen (depth, hw_flag, fullscreen_flag, font_file, font_size);
  b1 = create_buffer (0);
  b2 = create_buffer (100);
  b3 = create_buffer (10000);

  {
    struct buffer *bb[3] = { b1, b2, b3 };
    int  current = 1;
    int  i;

    buffer_connect (bb[current]);
    for (i=0; i<100; ++i) {
      int  next = jvrand_discrete (0, 3);
      int  j;

      screen_update ();
      if (current != next) {
	buffer_disconnect (bb[current]);
	current = next;
	buffer_connect (bb[current]);
      }

      for (j=0; j<1000; ++j) {
	struct charcell cc;

	cc.c = jvrand_discrete(0, 256);
	cc.attr = 0;
	cc.colors = JV_COLOR_PAIR(current+1, JV_COLOR_BLACK);
	buffer_putc (bb[current], cc);
      }
    }
  }

  delete_buffer (b1);
  delete_buffer (b2);
  delete_buffer (b3);
  close_screen ();

  SDL_Quit();

  return  0;
}

#endif
