/* client.c - infrastructure common to all clients
 *
 * A client is the collection of all resources connected to a VT.
 * Every instance owns a buffer for output and provides a class
 * method to receive input from the keyboard and from selections.
 *
 * Copyright (C) 2003, 2004  Jochen Voss.
 *
 * This program is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 2 of the License, or
 * (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 51 Franklin St, Fifth Floor, Boston, MA  02110-1301 USA
 *
 * $Id: client.c 6725 2005-11-09 23:15:58Z voss $
 */

#ifdef HAVE_CONFIG_H
#  include <config.h>
#endif

#include <stdlib.h>
#include <errno.h>

#include "base.h"
#include "ui.h"


struct client {
  const struct client_info *info; /* class method table */
  struct buffer *buf;		/* the associated buffer */
  void *data;			/* client data */
};

/* The currently active client, which receives keybord input
 * and has its buffer currently visible.  */
static struct client *current = NULL;

/**********************************************************************
 * selection handling
 */

static enum { cs_NORMAL, cs_POINT, cs_SELECT } select_state = cs_NORMAL;
static Uint16 *selection = NULL;

static void
save_selection (void)
/* Copy the selected text from the text buffer into the global
 * selection.  Unselect the text in the buffer.  */
{
  Uint16 *text = buffer_selection_get (current->buf);
  if (text) {
    xfree (selection);
    selection = text;
  }
  if (select_state != cs_NORMAL) {
    buffer_selection_stop (current->buf);
    select_state = cs_NORMAL;
  }
}

void
client_start_select (void)
/* Start the text selection mode.
 * Keyboard input is no longer forwarded to the buffer, now, but
 * used to control the selection instead.  */
{
  select_state = cs_POINT;
  buffer_selection_start (current->buf);
}

/**********************************************************************
 * global client functions
 */

struct client *
new_client (const struct client_info *ci, void *init_data, int backlog)
/* Create a new client.  If no other client is active make the new
 * client the active client.  */
{
  struct client *c;

  c = xnew (struct client);
  c->info = ci;
  c->buf = create_buffer (backlog);
  if (ci->init) {
    c->data = ci->init (c->buf, init_data);
  } else {
    c->data = init_data;
  }
  if (! current)  client_activate (c);

  return  c;
}

void
delete_client (struct client *c)
{
  if (current == c)  current = NULL;
  if (c->info->exit)  c->info->exit (c->data);
  delete_buffer (c->buf);
  
  xfree (c);
}

void
client_activate (struct client *c)
/* Activate a client.  If any other client is active, it is
 * deactivated and disconnected from the screen.  The current
 * selection is saved and the selection state reset to normal.  The
 * the client C is activated and has its buffer connected to the
 * screen.  */
{
  if (current) {
    if (select_state != cs_NORMAL)  save_selection ();
    if (current->info->deactivate)  current->info->deactivate (current->data);
    buffer_disconnect (current->buf);
  }
  current = c;
  select_state = cs_NORMAL;
  buffer_connect (current->buf);
  if (current->info->activate)  current->info->activate (current->data);
}

void *
client_apply (struct client *c, void *(*fn)(void *cdata, void *data),
	      void *data)
{
  return  fn (c->data, data);
}

static void
emit_char_cell_info (void)
{
  const struct buffer *b = current->buf;
  if (b->sel_point >= 0
      && b->sel_point < b->used_rows*b->cols) {
    int  cx = b->sel_point % b->cols;
    int  cy = b->sel_point / b->cols - b->pos;
    struct charcell  cc = screen_get_cell (cx, cy);
    char *astr = charcell_attr_name (cc.attr);
    message ("character at x=%d, y=%d: code U+%04X, %s, %s on %s",
	     cx, cy, cc.c, astr,
	     color_name(GET_COLOR_FG(cc.colors)),
	     color_name(GET_COLOR_BG(cc.colors)));
    xfree (astr);
  }
}

void
client_send_key (SDL_KeyboardEvent *key)
/* Process a key for the active client.
 * This handles the key codes for selection support and virtual scrolling.
 * Everything else is passed on to the client's class method 'send_key'.  */
{
  SDL_keysym *keysym = &(key->keysym);

  if (! current)  return;

  if (select_state == cs_POINT
      && keysym->mod&KMOD_SHIFT
      && (keysym->sym == SDLK_UP
	  || keysym->sym == SDLK_DOWN
	  || keysym->sym == SDLK_LEFT
	  || keysym->sym == SDLK_RIGHT)) {
    buffer_mark_set (current->buf);
    select_state = cs_SELECT;
  }
  if (select_state == cs_POINT || select_state == cs_SELECT) {
    switch (keysym->sym) {
    case SDLK_UP:
      buffer_point_up (current->buf);
      break;
    case SDLK_DOWN:
      buffer_point_down (current->buf);
      break;
    case SDLK_LEFT:
      buffer_point_left (current->buf);
      break;
    case SDLK_RIGHT:
      buffer_point_right (current->buf);
      break;
    case SDLK_SPACE:
      if (select_state == cs_POINT) {
	select_state = cs_SELECT;
	buffer_mark_set (current->buf);
      } else {
	save_selection ();
      }
      break;
    case SDLK_RETURN:
      save_selection ();
      if (selection) {
	ssize_t  res;
	res = client_send_string (selection);
	if (res < 0) {
	  error ("client_send_key: client_send_string failed: %m");
	}
      }
      break;
    case SDLK_ESCAPE:
      buffer_selection_stop (current->buf);
      select_state = cs_NORMAL;
      break;
    case SDLK_QUESTION:
    case SDLK_i:
      emit_char_cell_info ();
      break;
    default:
      break;
    }
  } else {			/* select_state == cs_NORMAL */
    if (keysym->sym == SDLK_PAGEUP && keysym->mod&KMOD_SHIFT) {
      buffer_vscroll_back (current->buf, 0);
    } else if (keysym->sym == SDLK_PAGEDOWN && keysym->mod&KMOD_SHIFT) {
      buffer_vscroll_forward (current->buf, 0);
    } else if (current->info->send_key) {
      current->info->send_key (key, current->data);
    }
  }
}

ssize_t
client_send_string (Uint16 *s)
/* Send an UCS-2 string to the client's input channel.  */
{
  size_t  len;
  Uint16 *ptr;
  
  if (! current || ! current->info->send_string) {
    errno = ENOTSUP;
    return  -1;
  }

  len = 0;
  ptr = s;
  while (*ptr++)  len += 2;
  
  return  current->info->send_string (s, len, current->data);
}

void
client_send_char (Uint16 c)
/* Send an individual UCS-2 character to the client's input channel.
 * TODO: error handling */
{
  if (! current)  return;
  if (current->info->send_string) {
    current->info->send_string (&c, 2, current->data);
  }
}
