/* control.c - implementation of the jvterm control center
 *
 * Copyright (C) 2003, 2005  Jochen Voss
 *
 * This program is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 2 of the License, or
 * (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 51 Franklin St, Fifth Floor, Boston, MA  02110-1301 USA
 *
 * $Id: control.c 6725 2005-11-09 23:15:58Z voss $
 */

#ifdef HAVE_CONFIG_H
#  include <config.h>
#endif

#include <stdio.h>

#include "base.h"
#include "backend.h"
#include "ui.h"


struct control_client_data {
  struct buffer *buf;
};

static void
control_print (struct control_client_data *ccd)
{
  char buffer[80];
  int  i;

  buffer_clear (ccd->buf);
  buffer_puts (ccd->buf, "jvterm, version " VERSION " (", 0, DEFAULT_COLOR);
  SDL_VideoDriverName (buffer, 80);
  buffer_puts (ccd->buf, buffer, 0, DEFAULT_COLOR);
  buffer_puts (ccd->buf, " video driver)\n\n", 0, DEFAULT_COLOR);

  buffer_puts (ccd->buf, "TERMINALS:\n", JV_ATTR_BRIGHT, DEFAULT_COLOR);
  buffer_puts (ccd->buf, "\n", 0, DEFAULT_COLOR);
  for (i=0; i<8; ++i) {
    int  attr = i==current_vt ? JV_ATTR_BRIGHT : 0;
    sprintf (buffer, "%c VT%02d", i==current_vt ? '*' : ' ', i+1);
    buffer_puts (ccd->buf, buffer, attr, DEFAULT_COLOR);
    if (vt[i]) {
      buffer_puts (ccd->buf, " ACTIVE\n", attr, DEFAULT_COLOR);
    } else {
      buffer_puts (ccd->buf, " -\n", attr, DEFAULT_COLOR);
    }
  }
  buffer_puts (ccd->buf, "\n", 0, DEFAULT_COLOR);

  buffer_puts (ccd->buf, "KEYS:\n", JV_ATTR_BRIGHT, DEFAULT_COLOR);
  buffer_puts (ccd->buf, "\n", 0, DEFAULT_COLOR);
  buffer_puts (ccd->buf, "  Left, Right - select virtual terminal\n",
	       0, DEFAULT_COLOR);
  buffer_puts (ccd->buf, "  RETURN - select active terminal or activate a new one\n",
	       0, DEFAULT_COLOR);
  buffer_puts (ccd->buf, "  S-RETURN - select existing terminal or open new one in meta mode\n",
	       0, DEFAULT_COLOR);
  buffer_puts (ccd->buf, "  q, ESC - return to default terminal\n",
	       0, DEFAULT_COLOR);
  buffer_puts (ccd->buf, "  Q - kill all clients and quit jvterm\n",
	       0, DEFAULT_COLOR);
  buffer_puts (ccd->buf, "  l - show message log\n", 0, DEFAULT_COLOR);
  buffer_puts (ccd->buf, "  i - iconify jvterm\n", 0, DEFAULT_COLOR);
}

/**********************************************************************
 * interface functions
 */

static void *
control_init (struct buffer *buf, void *init_data)
{
  struct control_client_data *ccd;

  ccd = xnew (struct control_client_data);
  ccd->buf = buf;
  
  return  ccd;
}

static void
control_exit (void *data)
{
  struct control_client_data *ccd = data;
  xfree (ccd);
}

static void
control_activate (void *data)
{
  struct control_client_data *ccd = data;
  control_print (ccd);
}

static void
control_send_key (SDL_KeyboardEvent *key, void *data)
{
  struct control_client_data *ccd = data;
  SDL_keysym *keysym = &(key->keysym);

  if (keysym->sym == SDLK_q && keysym->mod&KMOD_SHIFT) {
    shutdown_flag = 1;
    return;
  }

  switch (keysym->sym) {
  case SDLK_RETURN:
    select_vt (current_vt, keysym->mod&KMOD_SHIFT ? 2 : 1);
    break;
  case SDLK_LCTRL:
  case SDLK_RCTRL:
  case SDLK_ESCAPE:
  case SDLK_q:
    select_vt (vt[current_vt] ? current_vt : -1, 0);
    break;
  case SDLK_F1:
  case SDLK_F2:
  case SDLK_F3:
  case SDLK_F4:
  case SDLK_F5:
  case SDLK_F6:
  case SDLK_F7:
  case SDLK_F8:
  case SDLK_F12:
    select_vt (keysym->sym - SDLK_F1, 1);
    break;
  case SDLK_1:
  case SDLK_2:
  case SDLK_3:
  case SDLK_4:
  case SDLK_5:
  case SDLK_6:
  case SDLK_7:
  case SDLK_8:
    select_vt (keysym->sym - SDLK_1, 1);
    break;
  case SDLK_k:
    client_activate (keyclient);
    break;
  case SDLK_l:
  case SDLK_m:
    client_activate (messages);
    break;
  case SDLK_i:
    select_vt (vt[current_vt] ? current_vt : -1, 0);
    if (! SDL_WM_IconifyWindow ()) {
      error ("cannot iconify window");
    }
    break;
  case SDLK_DOWN:
  case SDLK_RIGHT:
    if (current_vt<7) {
      ++current_vt;
      control_print (ccd);
    }
    break;
  case SDLK_UP:
  case SDLK_LEFT:
    if (current_vt>0) {
      --current_vt;
      control_print (ccd);
    }
    break;
  default:
    break;
  }  
}

struct client_info control_client = {
  init: control_init,
  exit: control_exit,
  activate: control_activate,
  deactivate: NULL,
  send_key: control_send_key,
  send_string: NULL
};
