/* convert.c - convert between character encodings
 *
 * Copyright (C) 2003  Jochen Voss.
 *
 * This program is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 2 of the License, or
 * (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 51 Franklin St, Fifth Floor, Boston, MA  02110-1301 USA
 *
 * $Id: converter.c 6725 2005-11-09 23:15:58Z voss $
 */

#ifdef HAVE_CONFIG_H
#  include <config.h>
#endif

#include <stdio.h>
#include <stddef.h>
#include <stdlib.h>
#include <string.h>
#include <iconv.h>
#include <assert.h>
#include <errno.h>


#include "base.h"


#define INBUF_DEFAULT_SIZE 512
#define OUTBUF_DEFAULT_SIZE 2048


struct converter {
  iconv_t  c;			/* libc handle for the converter */

  const void *repl;		/* replacement character */
  size_t repl_len;

  write_fn  write;		/* function to emit the result */
  void *client_data;

  size_t  inbuf_used, inbuf_size; /* buffer for the input data */
  char *inbuf;

  size_t  outbuf_used, outbuf_size; /* buffer for the converted data */
  char *outbuf;

  char *id, *from, *to;		/* converter ID for debugging purposes */

  int  resync_utf8_flag;	/* 1 = we are currently dropping bytes from
				 *     an invalid UTF8 sequence */
};

/**********************************************************************
 * auxiliary functions
 */

static void
converter_append_input (struct converter *conv, const void *data, size_t len)
/* Append a chunk of data to the input buffer.  This will be queued
 * for processing with the 'iconv' function.  If necessary the input
 * buffer size is increased until the new data fits into the
 * buffer.  */
{
  if (conv->inbuf_used + len > conv->inbuf_size) {
    conv->inbuf_size = max (2*conv->inbuf_size, conv->inbuf_used + len);
    conv->inbuf = xrealloc (conv->inbuf, conv->inbuf_size);
  }
  memcpy (conv->inbuf+conv->inbuf_used, data, len);
  conv->inbuf_used += len;
}

static void
converter_append_output (struct converter *conv, const void *data, size_t len)
/* Append DATA to the output buffer.  This will be queued until it is
 * written out via the 'converter_flush' function.  If necessary the
 * output buffer size is increased until the new data fits into the
 * buffer.  */
{
  if (conv->outbuf_used + len > conv->outbuf_size) {
    conv->outbuf_size = max (2*conv->outbuf_size, conv->outbuf_used + len);
    conv->outbuf = xrealloc (conv->outbuf, conv->outbuf_size);
  }
  memcpy (conv->outbuf+conv->outbuf_used, data, len);
  conv->outbuf_used += len;
}

static void
converter_flush (struct converter *conv)
/* Emit as much data as possible from the output buffer using the
 * converter's 'write' function. */
{
  ssize_t  rc;

  if (conv->outbuf_used == 0)  return;
  
 retry_write:
  rc = conv->write (conv->outbuf, conv->outbuf_used, conv->client_data);
  if (rc == (ssize_t)-1) {
    if (errno == EINTR)  goto retry_write;
    if (errno != EAGAIN)  error ("%s write failed: %m", conv->id);
    return;
  } else {
    conv->outbuf_used -= rc;
    memmove (conv->outbuf, conv->outbuf+rc, conv->outbuf_used);
  }

  if (conv->outbuf_size > OUTBUF_DEFAULT_SIZE && conv->outbuf_used < 16) {
    conv->outbuf_size = OUTBUF_DEFAULT_SIZE;
    conv->outbuf = xrealloc (conv->outbuf, conv->outbuf_size);
  }
}

static int
utf8_is_continuation (unsigned char c)
/* Return true iff 'c' is an UTF8 continuation byte.  */
{
  return  (c >= 0x80 && c < 0xC0);
}

/**********************************************************************
 * public interface
 */

struct converter *
open_converter(const char *id,
	       const char *from, const char *to,
	       const void *repl, size_t repl_len,
	       write_fn write, void *client_data)
/* Create a new character enconding converter named ID.
 * The new converter will convert from character encoding FROM
 * into character encoding TO.  On invalid input it will emit
 * the replacement string REPL, which is of size REPL_LEN.
 * The converted data will be fed into the function WRITE.
 *
 * The ID argument is only used for debugging output.
 *
 * Return NULL on error and a handle to the new converter else.  */
{
  struct converter *conv;

  conv = xnew (struct converter);

  conv->c = iconv_open (to, from);
  if (conv->c == (iconv_t) -1) {
    error ("can't allocate converter: %m");
    xfree (conv);
    return  NULL;
  }
  conv->repl = repl;
  conv->repl_len = repl_len;
  conv->write = write;
  conv->client_data = client_data;

  conv->inbuf_used = 0;
  conv->inbuf_size = INBUF_DEFAULT_SIZE;
  conv->inbuf = xmalloc (conv->inbuf_size);

  conv->outbuf_used = 0;
  conv->outbuf_size = OUTBUF_DEFAULT_SIZE;
  conv->outbuf = xmalloc (conv->outbuf_size);

  conv->id = xstrdup (id);
  conv->from = xstrdup (from);
  conv->to = xstrdup (to);

  conv->resync_utf8_flag = 0;
  
  return  conv;
}

void
close_converter (struct converter *conv)
/* Free the converter CONV and all associated resources.  */
{
  converter_flush (conv);
  
  if (iconv_close (conv->c) != 0) {
    error ("close_converter (%s): close failed: %m", conv->id);
  }
  xfree (conv->id);
  xfree (conv->to);
  xfree (conv->from);
  xfree (conv->outbuf);
  xfree (conv->inbuf);
  xfree (conv);
}

void
converter_reset (struct converter *conv)
/* Add code into the output buffer to reset the conversion to the
 * initial state.  This function does not cause any output and thus
 * can be called from within converter write functions like
 * 'cooked_putchar' in "console.c".  */
{
  char *outptr;
  size_t  outbytesleft;
  size_t  rc;
  int  retry_flag;

  do {
    retry_flag = 0;
    outptr = conv->outbuf + conv->outbuf_used;
    outbytesleft = conv->outbuf_size - conv->outbuf_used;
    rc = iconv (conv->c, NULL, NULL, &outptr, &outbytesleft);
    conv->outbuf_used = outptr - conv->outbuf;
    
    if (rc == (size_t)-1) {
      if (errno == E2BIG) {
	conv->outbuf_size += 16;
	conv->outbuf = xrealloc (conv->outbuf, conv->outbuf_size);
	retry_flag = 1;
      } else {
	error ("converter_copy (%s): iconv reset failed: %m", conv->id);
      }
    }
  } while (retry_flag);
}

static int
converter_convert (struct converter *conv)
/* Convert data from inbuf and store the result in outbuf.
 * Return 1, iff there is more data to process, and 0 else.  */
{
  char *inptr, *outptr;
  size_t  inbytesleft, outbytesleft;
  size_t  rc;
  int  need_repl = 0;
  int  res = 1;

  inptr = conv->inbuf;
  inbytesleft = conv->inbuf_used;
  outptr = conv->outbuf + conv->outbuf_used;
  outbytesleft = conv->outbuf_size - conv->outbuf_used;

  if (conv->resync_utf8_flag) {
    while (inbytesleft>0 && utf8_is_continuation(*inptr)) {
      warning ("converter_convert (%s): dropping continuation byte 0x%02X",
	       conv->id, *((unsigned char *)inptr));
      ++inptr;
      --inbytesleft;
    }
    if (inbytesleft>0)  conv->resync_utf8_flag = 0;
  }
  
  rc = iconv (conv->c, &inptr, &inbytesleft, &outptr, &outbytesleft);
  if (rc == (size_t)-1) {
    switch (errno) {
    case EILSEQ:		/* invalid byte sequence */
      warning ("converter_convert (%s): dropping unexpected input byte 0x%02X",
	       conv->id, *((unsigned char *)inptr));
      need_repl = 1;
      if (strcmp(conv->from, "UTF-8") == 0
	  && ! utf8_is_continuation (*inptr)) {
	conv->resync_utf8_flag = 1;
      }
      assert (inbytesleft>0);
      ++inptr;
      --inbytesleft;
      break;
    case EINVAL:		/* incomplete input */
      res = 0;
      break;
    case E2BIG:			/* output buffer full */
      break;
    default:
      error ("converter_copy (%s): iconv failed: %m", conv->id);
      /* TODO: how to handle this? */
      res = 0;
      break;
    }
  }

  memmove (conv->inbuf, inptr, inbytesleft);
  conv->inbuf_used = inbytesleft;
  conv->outbuf_used = outptr - conv->outbuf;

  if (conv->inbuf_size > INBUF_DEFAULT_SIZE && conv->inbuf_used < 16) {
    conv->inbuf_size = INBUF_DEFAULT_SIZE;
    conv->inbuf = xrealloc (conv->inbuf, conv->inbuf_size);
  }
  
  if (need_repl) {
    converter_reset (conv);
    converter_append_output (conv, conv->repl, conv->repl_len);
  }

  if (conv->inbuf_used == 0)  res = 0;
  
  return  res;
}

void
converter_write (struct converter *conv, const void *buffer, size_t size)
/* Convert the data from a buffer.
 * The input data is stored in BUFFER and has length SIZE.  */
{
  char  try_convert;

  converter_append_input (conv, buffer, size);

  try_convert = (conv->inbuf_used > 0);
  while (try_convert || conv->outbuf_used > 0) {
    if (try_convert)  try_convert = converter_convert (conv);
    if (conv->outbuf_used > 0)  converter_flush (conv);
  }
}

ssize_t
converter_copy (struct converter *conv, int fd)
/* Read all available data from file descriptor FD and feed it into
 * the converter.  Return the number of transferred bytes or -1 on
 * error.  */
{
  size_t  bytes_read;
  char  try_read, try_convert;

  bytes_read = 0;

  try_read = 1;
  try_convert = (conv->inbuf_used > 0);

  while (try_read || try_convert || conv->outbuf_used > 0) {
    if (try_read) {
      ssize_t  rc;

    retry_read:
      rc = read (fd, conv->inbuf+conv->inbuf_used,
		 conv->inbuf_size-conv->inbuf_used);
      if (rc == (size_t)-1) {
	if (errno == EINTR)  goto retry_read;
	if (errno == EAGAIN) {
	  rc = 0;
	  try_read = 0;
	  try_convert = 0;
	} else {
	  if (errno != EIO) {
	    error ("converter_copy (%s): read failed: %m", conv->id);
	  }
	  return  -1;
	}
      } else {
	if (rc < conv->inbuf_size-conv->inbuf_used) try_read = 0;
	if (rc > 0) try_convert = 1;
	conv->inbuf_used += rc;
	bytes_read += rc;
      }
    }

    if (try_convert)  try_convert = converter_convert (conv);
    if (conv->outbuf_used > 0)  converter_flush (conv);
  }
  
  return  bytes_read;
}

/**********************************************************************
 * unit testing
 */

#ifdef TEST_CONVERTER

#include <stdint.h>

static ssize_t
emit_ucs2 (const void *buffer, size_t size, void *client_data)
{
  const uint16_t *output = buffer;
  int  i;

  assert (size % sizeof(uint16_t) == 0);

  putchar ('\n');
  printf ("result:");

  for (i=0; i<size/sizeof(uint16_t); ++i) {
    printf (" %u", output[i]);
  }
  putchar ('\n');

  return  size;
}

int
main ()
{
  const uint16_t unicode_replacement_character = 0xFFFD;
  struct converter *conv;
  uint8_t input1[] = {		/* UTF-8 for 32 <error> 3000 (truncated) ... */
    32, 128, 224, 174,
  };
  uint8_t input2[] = {		/* ... rest of 3000, then 32 */
    184, 32
  };

  conv = open_converter ("test", "UTF-8", "UCS-2",
			 &unicode_replacement_character, 2,
			 emit_ucs2, NULL);

  converter_write (conv, input1, sizeof (input1));
  converter_write (conv, input2, sizeof (input2));

  close_converter (conv);
  
  return  0;
}

#endif
