/* font.c - load and cache character glyphs
 *
 * This module loads a TTF font file and places the character glyphs
 * into SDL_Surfaces.  The ACS graphics characters are generated
 * manually and not taken from the font.
 *
 * Copyright (C) 2003, 2005  Jochen Voss
 *
 * This program is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 2 of the License, or
 * (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 51 Franklin St, Fifth Floor, Boston, MA  02110-1301 USA
 *
 * $Id: font.c 6725 2005-11-09 23:15:58Z voss $
 */

#ifdef HAVE_CONFIG_H
#  include <config.h>
#endif

#include <stdlib.h>
#include <stdarg.h>
#include <math.h>
#include <assert.h>

#include <ft2build.h>
#include FT_FREETYPE_H

#include <SDL.h>
#include <SDL_gfxPrimitives.h>

#include "base.h"
#include "backend.h"


#define  DEFAULT_PATH "/usr/share/fonts/truetype/ttf-bitstream-vera/"
#define  DEFAULT_FONT DEFAULT_PATH "VeraMono.ttf"


/**********************************************************************
 * a hash table to cache the glyph images
 */

#define NBUCKETS 389		/* smallest prime greater then 256+128 */

struct hash_entry {
  struct hash_entry *next;
  struct charcell  cc;		/* key */
  SDL_Surface *glyph;		/* payload */
};
struct hash_entry *glyph_table [NBUCKETS];

struct hash_info {
  unsigned  n, min, median, max;
};


static unsigned
key_fn (struct charcell cc)
{
  unsigned long  hash = (cc.colors<<24) + (cc.attr<<16) + cc.c;
  return  hash % NBUCKETS;
}

static int
hash_get_entry (struct charcell cc, struct hash_entry **res)
/* Get an entry from the hash table, adding one if none was present.
 * Return 1 iff an entry was found and 0 if a new one was allocated.  */
{
  struct hash_entry *ptr, *base;
  unsigned  key;

  key = key_fn (cc);
  base = ptr = glyph_table [key];
  while (ptr) {
    if (ptr->cc.c == cc.c
	&& ptr->cc.attr == cc.attr
	&& ptr->cc.colors == cc.colors) {
      *res = ptr;
      return  1;
    }
    ptr=ptr->next;
  }

  ptr = xnew (struct hash_entry);
  ptr->next = base;
  ptr->cc = cc;
  ptr->glyph = NULL;
  glyph_table [key] = ptr;
  
  *res = ptr;
  return  0;
}

static int
compare_integers (const void *pa, const void *pb)
/* A comparison function for use with 'qsort'.  */
{
  const int *a = (const int *) pa;
  const int *b = (const int *) pb;
  
  return (*a > *b) - (*a < *b);
}

static void
hash_analyse (struct hash_info *info)
/* Statistically analyse the hash table usage and place the result int INFO.
 * The output may be useful for performance tuning.  */
{
  unsigned  count [NBUCKETS], total;
  int  i;

  total = 0;
  for (i=0; i<NBUCKETS; ++i) {
    struct hash_entry *ptr;
    
    count[i] = 0;
    ptr = glyph_table[i];
    while (ptr) {
      ++ total;
      ++ count[i];
      ptr = ptr->next;
    }
  }
  qsort (count, NBUCKETS, sizeof(unsigned), compare_integers);
  info->n = total;
  info->min = count [0];
  info->max = count [NBUCKETS-1];
  info->median = count [NBUCKETS/2];
}

/**********************************************************************
 * load and handle the TTF fonts
 */

static  FT_Library  library;
static  FT_Face  face;

void
load_font (const char *ttf_file_name, double ptsize,
	   struct font_info *font_info)
/* Load a TrueType font into memory.
 *
 * If TTF_FILE_NAME is non-null, it is the file name to open.
 * Otherwise a default font is used.  The font is rendered at (design)
 * font size PTSIZE points.
 *
 * Carefully examine the available font geometry information and
 * the glyph sizes of 'M' and '(' to come up with a sensible
 * character cell size.
 *
 * Returns: information about the font geometry is stored in the
 * FONT_INFO fields. */
{
  int  cell_width, cell_height;
  double  ascent, descent;
  double  ulp, ult;
  double  m_horiAdvance, m_horiBearingX, m_width, m_height, m_horiBearingY;
  int  i, rc;

  if (! ttf_file_name)  ttf_file_name = DEFAULT_FONT;
  
  rc = FT_Init_FreeType (&library);
  if (rc)  fatal ("cannot init freetype library");

  message ("loading font %s", ttf_file_name);
  rc = FT_New_Face (library, ttf_file_name, 0, &face);
  if (rc == FT_Err_Unknown_File_Format) {
    fatal ("unsupport font file format");
  } else if (rc) {
    fatal ("cannot load font %s", ttf_file_name);
  }

  message ("choosing font size %gpt at 96 dpi", ptsize);
  rc = FT_Set_Char_Size (face, 0, (int)(ptsize*64.0+0.5), 96, 96);
  if (rc)  fatal ("cannot set font size %gpt", ptsize);

  rc = FT_Select_Charmap (face, FT_ENCODING_UNICODE);
  if (rc)  fatal ("cannot select unicode character map");
  
  message ("global font information:");
  cell_height = face->size->metrics.height / 64.0 + 0.5;
  ascent = face->size->metrics.ascender / 64.0;
  descent = face->size->metrics.descender / 64.0;
  message ("  baselineskip = %gpx", face->size->metrics.height / 64.0);
  message ("  ascent = %gpx", ascent);
  message ("  descent = %gpx", -descent);
  assert (descent <= 0);

  if (FT_IS_SCALABLE (face)) {
    ulp = FT_MulFix (face->underline_position,
		     face->size->metrics.y_scale) / 64.0;
    ult = FT_MulFix (face->underline_thickness,
		     face->size->metrics.y_scale) / 64.0;
  } else {
    ulp = floor(0.5*descent)+0.5;
    ult = 1;
  }
  message ("  underline_position = %gpx", ulp);
  message ("  underline_thickness = %gpx", ult);

  message ("per-character font information:");
  rc = FT_Load_Char (face, 'M', FT_LOAD_DEFAULT);
  if (rc)  fatal ("cannot access chracter 'M'");
  m_horiAdvance = face->glyph->metrics.horiAdvance / 64.0;
  m_horiBearingX = face->glyph->metrics.horiBearingX / 64.0;
  m_width = face->glyph->metrics.width / 64.0;
  m_height = face->glyph->metrics.height / 64.0;
  m_horiBearingY = face->glyph->metrics.horiBearingY / 64.0;
  message ("  the letter 'M': (%g+%g+%g)x(%g+%g)",
	   m_horiBearingX, m_width, m_horiAdvance-(m_horiBearingX+m_width),
	   m_horiBearingY, m_height-m_horiBearingY);

  cell_width = m_horiAdvance + 0.5;

  rc = FT_Load_Char (face, '(', FT_LOAD_DEFAULT);
  if (rc)  fatal ("cannot access chracter '('");
  m_horiAdvance = face->glyph->metrics.horiAdvance / 64.0;
  m_horiBearingX = face->glyph->metrics.horiBearingX / 64.0;
  m_width = face->glyph->metrics.width / 64.0;
  m_height = face->glyph->metrics.height / 64.0;
  m_horiBearingY = face->glyph->metrics.horiBearingY / 64.0;
  message ("  the letter '(': (%g+%g+%g)x(%g+%g)",
	   m_horiBearingX, m_width, m_horiAdvance-(m_horiBearingX+m_width),
	   m_horiBearingY, m_height-m_horiBearingY);

  if (cell_height > 1.5*m_height)  cell_height = 1.5*m_height+0.5;
  assert (m_horiBearingY>=0 && m_horiBearingY-m_height<=0);
  if (m_horiBearingY-m_height < descent)  descent = m_horiBearingY-m_height;
  if (m_horiBearingY > ascent)  ascent = m_horiBearingY;
  
  for (i=0; i<NBUCKETS; ++i)  glyph_table[i] = NULL;

  font_info->cell_width = cell_width;
  font_info->cell_height = cell_height;
  font_info->ascent = ascent*cell_height/(ascent-descent);
  font_info->descent = cell_height - font_info->ascent;
  message ("choosing cell dimension %dx(%d+%d)",
	   cell_width, font_info->ascent, font_info->descent);

  if (ult < 0.5)  ult = 0.5;
  if (ulp > -0.5*ult)  ulp = -0.5*ult;
  ulp += font_info->descent;
  if (ulp < 0.5*ult)  ulp = 0.5*ult;
  font_info->ula = cell_height - (ulp+0.5*ult);
  font_info->ulb = cell_height - (ulp-0.5*ult);
  message ("choosing underline position %.1f-%.1f",
	   font_info->ula, font_info->ulb);
}

void
unload_font (void)
/* Unload the allocated TTF font and free all the data from the
 * character glyph hash table.  */
{
  int  rc, i;

  {
    struct hash_info  info;
    hash_analyse (&info);
    message ("character glyph hash table usage:");
    message ("  glyph_hash.n = %d", info.n);
    message ("  glyph_hash.min = %d", info.min);
    message ("  glyph_hash.median = %d", info.median);
    message ("  glyph_hash.max = %d", info.max);
  }
  
  rc = FT_Done_FreeType (library);
  if (rc)  error ("cannot deinitialise freetype library");

  for (i=0; i<NBUCKETS; ++i) {
    if (glyph_table[i]) {
      struct hash_entry *ptr = glyph_table[i];
      glyph_table[i] = NULL;
      while (ptr) {
	struct hash_entry *next = ptr->next;
	SDL_FreeSurface (ptr->glyph);
	xfree (ptr);
	ptr = next;
      }
    }
  }
}

/**********************************************************************
 * auxiliary functions for glyph rendering
 */

static void
draw_rectangle (struct font_info *fi, unsigned char *data,
		double x0, double y0, double x1, double y1,
		int mflag)
/* Render a rectangle in the bitmap at DATA.  DATA has 1 byte per
 * pixel, fi->cell_width columns, and fi->cell_height rows.  The
 * rectangles lower left corner is (x0,y0), the upper right corner is
 * (x1,y1).  Horizontal and vertical coordinates range from 0.0 to
 * 1.0, where 0.0 is left resp. bottom.  For the rectangles margins
 * antialiasing is used.  If mflag is set, the new pixel values
 * replace the old ones, if mflag is unset, the pixel values are added
 * to the old values.  */
{
  int  w, h, i0, i1, j0, j1, i, j;
  
  w = fi->cell_width;
  h = fi->cell_height;

  i0 = floor (w*x0) + 0.5;
  i1 = ceil (w*x1) + 0.5;
  assert (i1>i0);
  j0 = floor (h*y0) + 0.5;
  j1 = ceil (h*y1) + 0.5;
  assert (j1>j0);

  for (i=i0; i<i1; ++i) {
    double qx;

    if (i==i0 && i==i1-1) {
      qx = 1 - w*(x1-x0);
    } else if (i == i0) {
      qx = 1 - (w*x0-i0);
    } else if (i == i1-1) {
      qx = 1 - (i1-w*x1);
    } else {
      qx = 1;
    }

    for (j=j0; j<j1; ++j) {
      double qy;
      unsigned char val;

      if (j==j0 && j==j1-1) {
	qy = 1 - h*(y1-y0);
      } else if (j == j0) {
	qy = 1 - (h*y0-j0);
      } else if (j == j1-1) {
	qy = 1 - (j1-h*y1);
      } else {
	qy = 1;
      }

      val = 255.99*qx*qy;
      if (mflag) {
	if (val > data[i+w*j]) data[i+w*j] = val;
      } else {
	data[i+w*j] += val;
      }
    }
  }
}

static SDL_Surface *
convert_to_surface (unsigned char *data, int w, int h, int pitch,
		    int x_offs, int y_offs, struct font_info *fi,
		    SDL_Color *forecol, SDL_Color *backcol)
/* Convert a grayscale pixel data buffer into a SDL surface.
 * The input data is an array at DATA, one byte per pixel.
 * The encoded image has width W, height H, and the start
 * bytes of adjacent rows are PITCH positions apart.
 *
 * The return value is a newly allocated SDL surface of the
 * cell size given by the font information structure FI,
 * it is rendered using the foreground color FORECOL and
 * the background color BACKCOL.  */
{
  SDL_Surface *raw, *glyph;
  SDL_Color colors[256];
  const SDL_VideoInfo *info;
  SDL_Rect dst;
  int  i, rc;
  
  raw = SDL_CreateRGBSurfaceFrom (data, w, h, 8, pitch, 0, 0, 0, 0);
  if (! raw)  fatal_sdl ("cannot create glyph surface");
  for (i=0; i<256; ++i) {
    double  q = i/255.0;
    colors[i].r = q*(forecol->r)+(1-q)*(backcol->r);
    colors[i].g = q*(forecol->g)+(1-q)*(backcol->g);
    colors[i].b = q*(forecol->b)+(1-q)*(backcol->b);
  }
  SDL_SetPalette (raw, SDL_LOGPAL, colors, 0, 256);

  info = SDL_GetVideoInfo ();
  glyph = SDL_CreateRGBSurface (SDL_SWSURFACE, fi->cell_width, fi->cell_height,
				info->vfmt->BitsPerPixel,
				info->vfmt->Rmask, info->vfmt->Gmask,
				info->vfmt->Bmask, info->vfmt->Amask);
  if (! glyph)  fatal_sdl ("cannot create glyph surface");
  SDL_FillRect (glyph, NULL, SDL_MapRGB (glyph->format,
					 backcol->r, backcol->g, backcol->b));
  
  dst.x = x_offs;
  dst.y = y_offs;
  rc = SDL_BlitSurface (raw, NULL, glyph, &dst);
  if (rc<0) {
    error_sdl ("cannot copy glyph");
    SDL_FreeSurface (glyph);
    glyph = NULL;
  }
  SDL_FreeSurface (raw);
  
  return  glyph;
}

static void
underline (SDL_Surface *glyph, SDL_Color *forecol, struct font_info *fi)
{
  int  min, max;

  min = floor(fi->ula+0.05);
  max = ceil(fi->ulb-0.05);
  if (max <= min) {
    double  p;
    p = fi->ulb - fi->ula;
    if (p>1)  p = 1;

    hlineRGBA (glyph, 0, glyph->w-1, min,
	       forecol->r, forecol->g, forecol->b, p*255);
  } else {
    double  p, q;
    int  i;
    
    p = 1 - (max - fi->ulb);
    q = 1 - (fi->ula - min);
    hlineRGBA (glyph, 0, glyph->w-1, min,
	       forecol->r, forecol->g, forecol->b, q*255);
    for (i=min+1; i<max-1; ++i)
      hlineRGBA (glyph, 0, glyph->w-1, i,
		 forecol->r, forecol->g, forecol->b, 255);
    hlineRGBA (glyph, 0, glyph->w-1, max-1,
	       forecol->r, forecol->g, forecol->b, p*255);
  }
}

static void
overlay (SDL_Surface *glyph, unsigned char *data,
	 int w, int h, int pitch, int x_offs, int y_offs, SDL_Color *col)
/* Modify glyph by overlaying DATA.  */
{
  SDL_PixelFormat  fmt;
  SDL_Surface *edit;
  Uint32 *pixels;
  int  rowstride;
  int  x, y;
  
  fmt.palette = NULL;
  fmt.BitsPerPixel = 32;
  fmt.BytesPerPixel = 4;
  fmt.Rloss = 0;
  fmt.Gloss = 0;
  fmt.Bloss = 0;
  fmt.Aloss = 8;
  fmt.Rshift = 24;
  fmt.Gshift = 16;
  fmt.Bshift = 8;
  fmt.Ashift = 0;
  fmt.Rmask = 0xFF000000;
  fmt.Gmask = 0x00FF0000;
  fmt.Bmask = 0x0000FF00;
  fmt.Amask = 0;
  fmt.colorkey = 0;
  fmt.alpha = 0;

  edit = SDL_ConvertSurface (glyph, &fmt, SDL_SWSURFACE);
  if (! edit) {
    error_sdl ("cannot generate overlay glyph");
    return;
  }
  pixels = edit->pixels;
  rowstride = edit->pitch / 4;

  SDL_LockSurface (edit);
  for (y=0; y<h; ++y) {
    for (x=0; x<w; ++x) {
      int  idx = (x+x_offs)+(y+y_offs)*rowstride;
      Uint8  tr = (pixels[idx]&fmt.Rmask) >> fmt.Rshift;
      Uint8  tg = (pixels[idx]&fmt.Gmask) >> fmt.Gshift;
      Uint8  tb = (pixels[idx]&fmt.Bmask) >> fmt.Bshift;
      double  val = data[x+y*pitch]/255.0;

      tr = val*col->r+(1-val)*tr+0.5;
      tg = val*col->g+(1-val)*tg+0.5;
      tb = val*col->b+(1-val)*tb+0.5;
      pixels[idx] = tr<<fmt.Rshift | tg<<fmt.Gshift | tb<<fmt.Bshift;
    }
  }
  SDL_UnlockSurface (edit);

  SDL_BlitSurface (edit, NULL, glyph, NULL);
  SDL_FreeSurface (edit);
}

/**********************************************************************
 * render character glyphs into SDL surfaces
 */

static SDL_Surface *
font_render_ft_glyph (Uint16 c, struct font_info *fi,
		      SDL_Color *forecol, SDL_Color *backcol)
/* Render character C from the loaded font into a newly allocated
 * SDL surface.  Return NULL if no glyph is available.  */
{
  int  rc, x_offs, y_offs;
  
  rc = FT_Load_Char (face, c, FT_LOAD_RENDER);
  if (rc)  return  NULL;

  x_offs = face->glyph->bitmap_left;
  x_offs += 0.5*(fi->cell_width - face->glyph->metrics.horiAdvance/64.0);
  y_offs = fi->cell_height - (face->glyph->bitmap_top + fi->descent);

  return  convert_to_surface (face->glyph->bitmap.buffer,
			      face->glyph->bitmap.width,
			      face->glyph->bitmap.rows,
			      face->glyph->bitmap.pitch,
			      x_offs, y_offs, fi, forecol, backcol);
}

static SDL_Surface *
font_render_shade (Uint16 c, struct font_info *fi,
		   SDL_Color *forecol, SDL_Color *backcol, double shade)
/* Render a character glyph which consists only of a full block
 * of gray.  SHADE==0 means black, SHADE==1 means white */
{
  int  w = fi->cell_width;
  int  h = fi->cell_height;
  unsigned char  data [w*h];
  int val;
  int  i;

  val = 256.0*shade;
  if (val>255)  val = 255;
  for (i=0; i<w*h; ++i)  data[i] = val;
  return  convert_to_surface (data, w, h, w, 0, 0, fi,
			      forecol, backcol);
}

static SDL_Surface *
font_render_box_char (Uint16 c, struct font_info *fi,
		      SDL_Color *forecol, SDL_Color *backcol,
		      double x0, double y0, double x1, double y1)
/* Render a character glyph which consists only of a rectangle with
 * lower left corner (X0,Y0) and the upper right corner (X1,Y1).
 * Horizontal and vertical coordinates range from 0.0 to 1.0, where
 * 0.0 is left resp. bottom.  */
{
  int  w = fi->cell_width;
  int  h = fi->cell_height;
  unsigned char  data [w*h];
  int  i;

  for (i=0; i<w*h; ++i)  data[i] = 0;
  draw_rectangle (fi, data, x0, y0, x1, y1, 0);

  return  convert_to_surface (data, w, h, w, 0, 0, fi,
			      forecol, backcol);
}

static SDL_Surface *
font_render_nbox_char (Uint16 c, struct font_info *fi,
		       SDL_Color *forecol, SDL_Color *backcol, int ul,
		       int n, ...)
/* Render a character glyph which consists only of a collection of
 * rectangles.  Horizontal and vertical coordinates range from 0.0 to
 * 1.0, where 0.0 is left resp. bottom.  */
{
  int  w = fi->cell_width;
  int  h = fi->cell_height;
  unsigned char  data [w*h];
  va_list args;
  double  x0, y0, x1, y1;
  int  i;

  va_start (args, n);
  for (i=0; i<w*h; ++i)  data[i] = 0;
  
  for (i=0; i<n; ++i) {
    x0 = va_arg (args, double);
    y0 = va_arg (args, double);
    x1 = va_arg (args, double);
    y1 = va_arg (args, double);
    draw_rectangle (fi, data, x0, y0, x1, y1, 0);
  }

  return  convert_to_surface (data, w, h, w, 0, 0, fi, forecol, backcol);
}

enum line_pos { line_pos_none, line_pos_a, line_pos_b, line_pos_ab };
enum line_weight { line_weight_none=-1,
		   line_weight_light, line_weight_heavy, line_weight_double };

static SDL_Surface *
font_render_line_char (Uint16 c, struct font_info *fi,
		       SDL_Color *forecol, SDL_Color *backcol,
		       enum line_pos hpos, enum line_weight hweight,
		       enum line_pos vpos, enum line_weight vweight)
/* Render a box drawing character glyph.  */
{
  int  w = fi->cell_width;
  int  h = fi->cell_height;
  double  q = (double)w/h;
  unsigned char  data [w*h];
  double  wtable[3] = { 0.07, 0.125, 0.2 };
  double  dx, dy, x0, x1, y0, y1;
  int  i;

  for (i=0; i<w*h; ++i)  data[i] = 0;
  dx = wtable[vweight];
  dy = q*wtable[hweight];

  x0 = x1 = y0 = y1 = 0;
  switch (hpos) {
  case line_pos_a:
    x0 = 0;
    x1 = 0.5+dx;
    break;
  case line_pos_b:
    x0 = 0.5-dx;
    x1 = 1;
    break;
  case line_pos_ab:
    x0 = 0;
    x1 = 1;
    break;
  case line_pos_none:
    break;
  }
  switch (vpos) {
  case line_pos_a:
    y0 = 0;
    y1 = 0.5+dy;
    break;
  case line_pos_b:
    y0 = 0.5-dy;
    y1 = 1;
    break;
  case line_pos_ab:
    y0 = 0;
    y1 = 1;
    break;
  case line_pos_none:
    break;
  }
  
  if (hpos != line_pos_none) {
    if (hweight == line_weight_double) {
      draw_rectangle (fi, data, x0, 0.5-dy, x1, 0.5-dy/3, 1);
      draw_rectangle (fi, data, x0, 0.5+dy/3, x1, 0.5+dy, 1);
    } else {
      draw_rectangle (fi, data, x0, 0.5-dy, x1, 0.5+dy, 1);
    }
  }
  if (vpos != line_pos_none) {
    if (vweight == line_weight_double) {
      draw_rectangle (fi, data, 0.5-dx, y0, 0.5-dx/3, y1, 1);
      draw_rectangle (fi, data, 0.5+dx/3, y0, 0.5+dx, y1, 1);
    } else {
      draw_rectangle (fi, data, 0.5-dx, y0, 0.5+dx, y1, 1);
    }
  }

  return  convert_to_surface (data, w, h, w, 0, 0, fi, forecol, backcol);
}

static SDL_Surface *
font_render_cross (Uint16 c, struct font_info *fi,
		   SDL_Color *forecol, SDL_Color *backcol,
		   enum line_weight up, enum line_weight down,
		   enum line_weight left, enum line_weight right)
/* Render a box drawing character glyph.  These may be more ugly than
 * the characters rendered by 'font_render_line_char', but you can
 * give the strength for all four possible arms separately.  */
{
  int  w = fi->cell_width;
  int  h = fi->cell_height;
  double  q = (double)w/h;
  unsigned char  data [w*h];
  double  wtable[3] = { 0.07, 0.125, 0.2 };
  int  i;

  for (i=0; i<w*h; ++i)  data[i] = 0;
  if (up != line_weight_none) {
    draw_rectangle (fi, data,
		    0.5-wtable[up], 0, 0.5+wtable[up], 0.5, 1);
  }
  if (down != line_weight_none) {
    draw_rectangle (fi, data,
		    0.5-wtable[down], 0.5, 0.5+wtable[down], 1.0, 1);
  }
  if (left != line_weight_none) {
    draw_rectangle (fi, data,
		    0, 0.5-q*wtable[left], 0.5, 0.5+q*wtable[left], 1);
  }
  if (right != line_weight_none) {
    draw_rectangle (fi, data,
		    0.5, 0.5-q*wtable[right], 1.0, 0.5+q*wtable[right], 1);
  }

  return  convert_to_surface (data, w, h, w, 0, 0, fi, forecol, backcol);
}

/**********************************************************************
 * high level glyph handling
 */

static SDL_Surface *
font_create_unicode_glyph (Uint16 c, struct font_info *fi,
			   SDL_Color *forecol, SDL_Color *backcol)
/* Create a SDL surface containing a glyph for the UCS2 character c.
 * Return NULL if no glyph is available.  Several graphics characters
 * are rendered manually, the remaining glyphs are taken from the font
 * file.  */
{
  SDL_Surface *glyph = NULL;
  int  w = fi->cell_width;
  int  h = fi->cell_height;
  double  q = (double)w/h;
  
  switch (c) {
  case 0x203E:			/* OVERLINE */
    glyph = font_render_box_char (c, fi, forecol, backcol, 0, 0, 1, 1/16.0);
    break;
  case 0x2500:			/* BOX DRAWINGS LIGHT HORIZONTAL */
    glyph = font_render_line_char (c, fi, forecol, backcol,
				   line_pos_ab, line_weight_light,
				   line_pos_none, line_weight_light);
    break;
  case 0x2501:			/* BOX DRAWINGS HEAVY HORIZONTAL */
    glyph = font_render_line_char (c, fi, forecol, backcol,
				   line_pos_ab, line_weight_heavy,
				   line_pos_none, line_weight_heavy);
    break;
  case 0x2502:			/* BOX DRAWINGS LIGHT VERTICAL */
    glyph = font_render_line_char (c, fi, forecol, backcol,
				   line_pos_none, line_weight_light,
				   line_pos_ab, line_weight_light);
    break;
  case 0x2503:			/* BOX DRAWINGS HEAVY VERTICAL */
    glyph = font_render_line_char (c, fi, forecol, backcol,
				   line_pos_none, line_weight_heavy,
				   line_pos_ab, line_weight_heavy);
    break;
  case 0x2504:	       /* BOX DRAWINGS LIGHT TRIPLE DASH HORIZONTAL */
    glyph = font_render_nbox_char (c, fi, forecol, backcol, 3,
				   0.0, 0.5-q*2.5*0.14, 1.0, 0.5-q*1.5*0.14,
				   0.0, 0.5-q*0.5*0.14, 1.0, 0.5+q*0.5*0.14,
				   0.0, 0.5+q*1.5*0.14, 1.0, 0.5+q*2.5*0.14);
    break;
  case 0x2505:	       /* BOX DRAWINGS HEAVY TRIPLE DASH HORIZONTAL */
    glyph = font_render_nbox_char (c, fi, forecol, backcol, 3,
				   0.0, 0.5-q*2.5*0.25, 1.0, 0.5-q*1.5*0.25,
				   0.0, 0.5-q*0.5*0.25, 1.0, 0.5+q*0.5*0.25,
				   0.0, 0.5+q*1.5*0.25, 1.0, 0.5+q*2.5*0.25);
    break;
  case 0x2506:	       /* BOX DRAWINGS LIGHT TRIPLE DASH VERTICAL */
    glyph = font_render_nbox_char (c, fi, forecol, backcol, 3,
				   0.5-2.5*0.14, 0.0, 0.5-1.5*0.14, 1.0,
				   0.5-0.5*0.14, 0.0, 0.5+0.5*0.14, 1.0,
				   0.5+1.5*0.14, 0.0, 0.5+2.5*0.14, 1.0);
    break;
  case 0x2507:	       /* BOX DRAWINGS HEAVY TRIPLE DASH VERTICAL */
    glyph = font_render_nbox_char (c, fi, forecol, backcol, 3,
				   0.5-2.5*0.19, 0.0, 0.5-1.5*0.19, 1.0,
				   0.5-0.5*0.19, 0.0, 0.5+0.5*0.19, 1.0,
				   0.5+1.5*0.19, 0.0, 0.5+2.5*0.19, 1.0);
    break;
  case 0x2508:	    /* BOX DRAWINGS LIGHT QUADRUPLE DASH HORIZONTAL */
    glyph = font_render_nbox_char (c, fi, forecol, backcol, 4,
				   0.0, 0.5-q*3.5*0.14, 1.0, 0.5-q*2.5*0.14,
				   0.0, 0.5-q*1.5*0.14, 1.0, 0.5-q*0.5*0.14,
				   0.0, 0.5+q*0.5*0.14, 1.0, 0.5+q*1.5*0.14,
				   0.0, 0.5+q*2.5*0.14, 1.0, 0.5+q*3.5*0.14);
    break;
  case 0x2509:	    /* BOX DRAWINGS HEAVY QUADRUPLE DASH HORIZONTAL */
    glyph = font_render_nbox_char (c, fi, forecol, backcol, 4,
				   0.0, 0.5-q*3.5*0.25, 1.0, 0.5-q*2.5*0.25,
				   0.0, 0.5-q*1.5*0.25, 1.0, 0.5-q*0.5*0.25,
				   0.0, 0.5+q*0.5*0.25, 1.0, 0.5+q*1.5*0.25,
				   0.0, 0.5+q*2.5*0.25, 1.0, 0.5+q*3.5*0.25);
    break;
  case 0x250A:	       /* BOX DRAWINGS LIGHT QUADRUPLE DASH VERTICAL */
    glyph = font_render_nbox_char (c, fi, forecol, backcol, 4,
				   0.5-3.5*0.09, 0.0, 0.5-2.5*0.09, 1.0,
				   0.5-1.5*0.09, 0.0, 0.5-0.5*0.09, 1.0,
				   0.5+0.5*0.09, 0.0, 0.5+1.5*0.09, 1.0,
				   0.5+2.5*0.09, 0.0, 0.5+3.5*0.09, 1.0);
    break;
  case 0x250B:	       /* BOX DRAWINGS HEAVY QUADRUPLE DASH VERTICAL */
    glyph = font_render_nbox_char (c, fi, forecol, backcol, 4,
				   0.5-3.5*0.14, 0.0, 0.5-2.5*0.14, 1.0,
				   0.5-1.5*0.14, 0.0, 0.5-0.5*0.14, 1.0,
				   0.5+0.5*0.14, 0.0, 0.5+1.5*0.14, 1.0,
				   0.5+2.5*0.14, 0.0, 0.5+3.5*0.14, 1.0);
    break;
  case 0x250C:		       /* BOX DRAWINGS LIGHT DOWN AND RIGHT */
    glyph = font_render_line_char (c, fi, forecol, backcol,
				   line_pos_b, line_weight_light,
				   line_pos_b, line_weight_light);
    break;
  case 0x250D:		 /* BOX DRAWINGS DOWN LIGHT AND RIGHT HEAVY */
    glyph = font_render_line_char (c, fi, forecol, backcol,
				   line_pos_b, line_weight_heavy,
				   line_pos_b, line_weight_light);
    break;
  case 0x250E:		 /* BOX DRAWINGS DOWN HEAVY AND RIGHT LIGHT */
    glyph = font_render_line_char (c, fi, forecol, backcol,
				   line_pos_b, line_weight_light,
				   line_pos_b, line_weight_heavy);
    break;
  case 0x250F:		       /* BOX DRAWINGS HEAVY DOWN AND RIGHT */
    glyph = font_render_line_char (c, fi, forecol, backcol,
				   line_pos_b, line_weight_heavy,
				   line_pos_b, line_weight_heavy);
    break;
  case 0x2510:			/* BOX DRAWINGS LIGHT DOWN AND LEFT */
    glyph = font_render_line_char (c, fi, forecol, backcol,
				   line_pos_a, line_weight_light,
				   line_pos_b, line_weight_light);
    break;
  case 0x2511:		  /* BOX DRAWINGS DOWN LIGHT AND LEFT HEAVY */
    glyph = font_render_line_char (c, fi, forecol, backcol,
				   line_pos_a, line_weight_heavy,
				   line_pos_b, line_weight_light);
    break;
  case 0x2512:		  /* BOX DRAWINGS DOWN HEAVY AND LEFT LIGHT */
    glyph = font_render_line_char (c, fi, forecol, backcol,
				   line_pos_a, line_weight_light,
				   line_pos_b, line_weight_heavy);
    break;
  case 0x2513:			/* BOX DRAWINGS HEAVY DOWN AND LEFT */
    glyph = font_render_line_char (c, fi, forecol, backcol,
				   line_pos_a, line_weight_heavy,
				   line_pos_b, line_weight_heavy);
    break;
  case 0x2514:			/* BOX DRAWINGS LIGHT UP AND RIGHT */
    glyph = font_render_line_char (c, fi, forecol, backcol,
				   line_pos_b, line_weight_light,
				   line_pos_a, line_weight_light);
    break;
  case 0x2515:		   /* BOX DRAWINGS UP LIGHT AND RIGHT HEAVY */
    glyph = font_render_line_char (c, fi, forecol, backcol,
				   line_pos_b, line_weight_heavy,
				   line_pos_a, line_weight_light);
    break;
  case 0x2516:		   /* BOX DRAWINGS UP HEAVY AND RIGHT LIGHT */
    glyph = font_render_line_char (c, fi, forecol, backcol,
				   line_pos_b, line_weight_light,
				   line_pos_a, line_weight_heavy);
    break;
  case 0x2517:			/* BOX DRAWINGS HEAVY UP AND RIGHT */
    glyph = font_render_line_char (c, fi, forecol, backcol,
				   line_pos_b, line_weight_heavy,
				   line_pos_a, line_weight_heavy);
    break;
  case 0x2518:			/* BOX DRAWINGS LIGHT UP AND LEFT */
    glyph = font_render_line_char (c, fi, forecol, backcol,
				   line_pos_a, line_weight_light,
				   line_pos_a, line_weight_light);
    break;
  case 0x2519:		    /* BOX DRAWINGS UP LIGHT AND LEFT HEAVY */
    glyph = font_render_line_char (c, fi, forecol, backcol,
				   line_pos_a, line_weight_heavy,
				   line_pos_a, line_weight_light);
    break;
  case 0x251A:		    /* BOX DRAWINGS UP HEAVY AND LEFT LIGHT */
    glyph = font_render_line_char (c, fi, forecol, backcol,
				   line_pos_a, line_weight_light,
				   line_pos_a, line_weight_heavy);
    break;
  case 0x251B:			/* BOX DRAWINGS HEAVY UP AND LEFT */
    glyph = font_render_line_char (c, fi, forecol, backcol,
				   line_pos_a, line_weight_heavy,
				   line_pos_a, line_weight_heavy);
    break;
  case 0x251C:		   /* BOX DRAWINGS LIGHT VERTICAL AND RIGHT */
    glyph = font_render_line_char (c, fi, forecol, backcol,
				   line_pos_b, line_weight_light,
				   line_pos_ab, line_weight_light);
    break;
  case 0x251D:	     /* BOX DRAWINGS VERTICAL LIGHT AND RIGHT HEAVY */
    glyph = font_render_line_char (c, fi, forecol, backcol,
				   line_pos_b, line_weight_heavy,
				   line_pos_ab, line_weight_light);
    break;
  case 0x251E:	      /* BOX DRAWINGS UP HEAVY AND RIGHT DOWN LIGHT */
    glyph = font_render_cross (c, fi, forecol, backcol,
			       line_weight_heavy, line_weight_light,
			       line_weight_none, line_weight_light);
    break;
  case 0x251F:	      /* BOX DRAWINGS DOWN HEAVY AND RIGHT UP LIGHT */
    glyph = font_render_cross (c, fi, forecol, backcol,
			       line_weight_light, line_weight_heavy,
			       line_weight_none, line_weight_light);
    break;
  case 0x2520:	     /* BOX DRAWINGS VERTICAL HEAVY AND RIGHT LIGHT */
    glyph = font_render_line_char (c, fi, forecol, backcol,
				   line_pos_b, line_weight_light,
				   line_pos_ab, line_weight_heavy);
    break;
  case 0x2521:	      /* BOX DRAWINGS DOWN LIGHT AND RIGHT UP HEAVY */
    glyph = font_render_cross (c, fi, forecol, backcol,
			       line_weight_heavy, line_weight_light,
			       line_weight_none, line_weight_light);
    break;
  case 0x2522:	      /* BOX DRAWINGS UP LIGHT AND RIGHT DOWN HEAVY */
    glyph = font_render_cross (c, fi, forecol, backcol,
			       line_weight_light, line_weight_heavy,
			       line_weight_none, line_weight_heavy);
    break;
  case 0x2523:		   /* BOX DRAWINGS HEAVY VERTICAL AND RIGHT */
    glyph = font_render_line_char (c, fi, forecol, backcol,
				   line_pos_b, line_weight_heavy,
				   line_pos_ab, line_weight_heavy);
    break;
  case 0x2524:		    /* BOX DRAWINGS LIGHT VERTICAL AND LEFT */
    glyph = font_render_line_char (c, fi, forecol, backcol,
				   line_pos_a, line_weight_light,
				   line_pos_ab, line_weight_light);
    break;
  case 0x2525:	      /* BOX DRAWINGS VERTICAL LIGHT AND LEFT HEAVY */
    glyph = font_render_line_char (c, fi, forecol, backcol,
				   line_pos_a, line_weight_heavy,
				   line_pos_ab, line_weight_light);
    break;
  case 0x2526:	       /* BOX DRAWINGS UP HEAVY AND LEFT DOWN LIGHT */
    glyph = font_render_cross (c, fi, forecol, backcol,
			       line_weight_heavy, line_weight_light,
			       line_weight_light, line_weight_none);
    break;
  case 0x2527:	       /* BOX DRAWINGS DOWN HEAVY AND LEFT UP LIGHT */
    glyph = font_render_cross (c, fi, forecol, backcol,
			       line_weight_light, line_weight_heavy,
			       line_weight_light, line_weight_none);
    break;
  case 0x2528:	      /* BOX DRAWINGS VERTICAL HEAVY AND LEFT LIGHT */
    glyph = font_render_line_char (c, fi, forecol, backcol,
				   line_pos_a, line_weight_light,
				   line_pos_ab, line_weight_heavy);
    break;
  case 0x2529:	       /* BOX DRAWINGS DOWN LIGHT AND LEFT UP HEAVY */
    glyph = font_render_cross (c, fi, forecol, backcol,
			       line_weight_heavy, line_weight_light,
			       line_weight_heavy, line_weight_none);
    break;
  case 0x252A:	       /* BOX DRAWINGS UP LIGHT AND LEFT DOWN HEAVY */
    glyph = font_render_cross (c, fi, forecol, backcol,
			       line_weight_light, line_weight_heavy,
			       line_weight_heavy, line_weight_none);
    break;
  case 0x252B:		    /* BOX DRAWINGS HEAVY VERTICAL AND LEFT */
    glyph = font_render_line_char (c, fi, forecol, backcol,
				   line_pos_a, line_weight_heavy,
				   line_pos_ab, line_weight_heavy);
    break;
  case 0x252C:		  /* BOX DRAWINGS LIGHT DOWN AND HORIZONTAL */
    glyph = font_render_line_char (c, fi, forecol, backcol,
				   line_pos_ab, line_weight_light,
				   line_pos_b, line_weight_light);
    break;
  case 0x252D:	    /* BOX DRAWINGS LEFT HEAVY AND RIGHT DOWN LIGHT */
    glyph = font_render_cross (c, fi, forecol, backcol,
			       line_weight_none, line_weight_light,
			       line_weight_heavy, line_weight_light);
    break;
  case 0x252E:	    /* BOX DRAWINGS RIGHT HEAVY AND LEFT DOWN LIGHT */
    glyph = font_render_cross (c, fi, forecol, backcol,
			       line_weight_none, line_weight_light,
			       line_weight_light, line_weight_heavy);
    break;
  case 0x252F:	    /* BOX DRAWINGS DOWN LIGHT AND HORIZONTAL HEAVY */
    glyph = font_render_line_char (c, fi, forecol, backcol,
				   line_pos_ab, line_weight_heavy,
				   line_pos_b, line_weight_light);
    break;
  case 0x2530:	    /* BOX DRAWINGS DOWN HEAVY AND HORIZONTAL LIGHT */
    glyph = font_render_line_char (c, fi, forecol, backcol,
				   line_pos_ab, line_weight_light,
				   line_pos_b, line_weight_heavy);
    break;
  case 0x2531:	    /* BOX DRAWINGS RIGHT LIGHT AND LEFT DOWN HEAVY */
    glyph = font_render_cross (c, fi, forecol, backcol,
			       line_weight_none, line_weight_heavy,
			       line_weight_heavy, line_weight_light);
    break;
  case 0x2532:	    /* BOX DRAWINGS LEFT LIGHT AND RIGHT DOWN HEAVY */
    glyph = font_render_cross (c, fi, forecol, backcol,
			       line_weight_none, line_weight_heavy,
			       line_weight_light, line_weight_heavy);
    break;
  case 0x2533:		  /* BOX DRAWINGS HEAVY DOWN AND HORIZONTAL */
    glyph = font_render_line_char (c, fi, forecol, backcol,
				   line_pos_ab, line_weight_heavy,
				   line_pos_b, line_weight_heavy);
    break;
  case 0x2534:		    /* BOX DRAWINGS LIGHT UP AND HORIZONTAL */
    glyph = font_render_line_char (c, fi, forecol, backcol,
				   line_pos_ab, line_weight_light,
				   line_pos_a, line_weight_light);
    break;
  case 0x2535:	      /* BOX DRAWINGS LEFT HEAVY AND RIGHT UP LIGHT */
    glyph = font_render_cross (c, fi, forecol, backcol,
			       line_weight_light, line_weight_none,
			       line_weight_heavy, line_weight_light);
    break;
  case 0x2536:	      /* BOX DRAWINGS RIGHT HEAVY AND LEFT UP LIGHT */
    glyph = font_render_cross (c, fi, forecol, backcol,
			       line_weight_light, line_weight_none,
			       line_weight_light, line_weight_heavy);
    break;
  case 0x2537:	      /* BOX DRAWINGS UP LIGHT AND HORIZONTAL HEAVY */
    glyph = font_render_line_char (c, fi, forecol, backcol,
				   line_pos_ab, line_weight_heavy,
				   line_pos_a, line_weight_light);
    break;
  case 0x2538:	      /* BOX DRAWINGS UP HEAVY AND HORIZONTAL LIGHT */
    glyph = font_render_line_char (c, fi, forecol, backcol,
				   line_pos_ab, line_weight_light,
				   line_pos_a, line_weight_heavy);
    break;
  case 0x2539:	      /* BOX DRAWINGS RIGHT LIGHT AND LEFT UP HEAVY */
    glyph = font_render_cross (c, fi, forecol, backcol,
			       line_weight_heavy, line_weight_none,
			       line_weight_heavy, line_weight_light);
    break;
  case 0x253A:	      /* BOX DRAWINGS LEFT LIGHT AND RIGHT UP HEAVY */
    glyph = font_render_cross (c, fi, forecol, backcol,
			       line_weight_heavy, line_weight_none,
			       line_weight_light, line_weight_heavy);
    break;
  case 0x253B:		    /* BOX DRAWINGS HEAVY UP AND HORIZONTAL */
    glyph = font_render_line_char (c, fi, forecol, backcol,
				   line_pos_ab, line_weight_heavy,
				   line_pos_a, line_weight_heavy);
    break;
  case 0x253C:	      /* BOX DRAWINGS LIGHT VERTICAL AND HORIZONTAL */
    glyph = font_render_line_char (c, fi, forecol, backcol,
				   line_pos_ab, line_weight_light,
				   line_pos_ab, line_weight_light);
    break;
  case 0x253D:	/* BOX DRAWINGS LEFT HEAVY AND RIGHT VERTICAL LIGHT */
    glyph = font_render_cross (c, fi, forecol, backcol,
			       line_weight_light, line_weight_light,
			       line_weight_heavy, line_weight_light);
    break;
  case 0x253E:	/* BOX DRAWINGS RIGHT HEAVY AND LEFT VERTICAL LIGHT */
    glyph = font_render_cross (c, fi, forecol, backcol,
			       line_weight_light, line_weight_light,
			       line_weight_light, line_weight_heavy);
    break;
  case 0x253F:	/* BOX DRAWINGS VERTICAL LIGHT AND HORIZONTAL HEAVY */
    glyph = font_render_line_char (c, fi, forecol, backcol,
				   line_pos_ab, line_weight_heavy,
				   line_pos_ab, line_weight_light);
    break;
  case 0x2540:	 /* BOX DRAWINGS UP HEAVY AND DOWN HORIZONTAL LIGHT */
    glyph = font_render_cross (c, fi, forecol, backcol,
			       line_weight_heavy, line_weight_light,
			       line_weight_light, line_weight_light);
    break;
  case 0x2541:	 /* BOX DRAWINGS DOWN HEAVY AND UP HORIZONTAL LIGHT */
    glyph = font_render_cross (c, fi, forecol, backcol,
			       line_weight_light, line_weight_heavy,
			       line_weight_light, line_weight_light);
    break;
  case 0x2542:	/* BOX DRAWINGS VERTICAL HEAVY AND HORIZONTAL LIGHT */
    glyph = font_render_line_char (c, fi, forecol, backcol,
				   line_pos_ab, line_weight_light,
				   line_pos_ab, line_weight_heavy);
    break;
  case 0x2543:	 /* BOX DRAWINGS LEFT UP HEAVY AND RIGHT DOWN LIGHT */
    glyph = font_render_cross (c, fi, forecol, backcol,
			       line_weight_heavy, line_weight_light,
			       line_weight_heavy, line_weight_light);
    break;
  case 0x2544:	 /* BOX DRAWINGS RIGHT UP HEAVY AND LEFT DOWN LIGHT */
    glyph = font_render_cross (c, fi, forecol, backcol,
			       line_weight_heavy, line_weight_light,
			       line_weight_light, line_weight_heavy);
    break;
  case 0x2545:	/* BOX DRAWINGS LEFT DOWN HEAVY AND RIGHT UP LIGHT */
    glyph = font_render_cross (c, fi, forecol, backcol,
			       line_weight_light, line_weight_heavy,
			       line_weight_heavy, line_weight_light);
    break;
  case 0x2546:	/* BOX DRAWINGS RIGHT DOWN HEAVY AND LEFT UP LIGHT */
    glyph = font_render_cross (c, fi, forecol, backcol,
			       line_weight_light, line_weight_heavy,
			       line_weight_light, line_weight_heavy);
    break;
  case 0x2547:	 /* BOX DRAWINGS DOWN LIGHT AND UP HORIZONTAL HEAVY */
    glyph = font_render_cross (c, fi, forecol, backcol,
			       line_weight_heavy, line_weight_light,
			       line_weight_heavy, line_weight_heavy);
    break;
  case 0x2548:	 /* BOX DRAWINGS UP LIGHT AND DOWN HORIZONTAL HEAVY */
    glyph = font_render_cross (c, fi, forecol, backcol,
			       line_weight_light, line_weight_heavy,
			       line_weight_heavy, line_weight_heavy);
    break;
  case 0x2549:	/* BOX DRAWINGS RIGHT LIGHT AND LEFT VERTICAL HEAVY */
    glyph = font_render_cross (c, fi, forecol, backcol,
			       line_weight_heavy, line_weight_heavy,
			       line_weight_heavy, line_weight_light);
    break;
  case 0x254A:	/* BOX DRAWINGS LEFT LIGHT AND RIGHT VERTICAL HEAVY */
    glyph = font_render_cross (c, fi, forecol, backcol,
			       line_weight_heavy, line_weight_heavy,
			       line_weight_light, line_weight_heavy);
    break;
  case 0x254B:	      /* BOX DRAWINGS HEAVY VERTICAL AND HORIZONTAL */
    glyph = font_render_line_char (c, fi, forecol, backcol,
				   line_pos_ab, line_weight_heavy,
				   line_pos_ab, line_weight_heavy);
    break;
  case 0x254C:	       /* BOX DRAWINGS LIGHT DOUBLE DASH HORIZONTAL */
    glyph = font_render_nbox_char (c, fi, forecol, backcol, 2,
				   0.0, 0.5-q*1.5*0.14, 1.0, 0.5-q*0.5*0.14,
				   0.0, 0.5+q*0.5*0.14, 1.0, 0.5+q*1.5*0.14);
    break;
  case 0x254D:	       /* BOX DRAWINGS HEAVY DOUBLE DASH HORIZONTAL */
    glyph = font_render_nbox_char (c, fi, forecol, backcol, 2,
				   0.0, 0.5-q*1.5*0.25, 1.0, 0.5-q*0.5*0.25,
				   0.0, 0.5+q*0.5*0.25, 1.0, 0.5+q*1.5*0.25);
    break;
  case 0x254E:	       /* BOX DRAWINGS LIGHT DOUBLE DASH VERTICAL */
    glyph = font_render_nbox_char (c, fi, forecol, backcol, 2,
				   0.5-1.5*0.14, 0.0, 0.5-0.5*0.14, 1.0,
				   0.5+0.5*0.14, 0.0, 0.5+1.5*0.14, 1.0);
    break;
  case 0x254F:	       /* BOX DRAWINGS HEAVY DOUBLE DASH VERTICAL */
    glyph = font_render_nbox_char (c, fi, forecol, backcol, 2,
				   0.5-1.5*0.25, 0.0, 0.5-0.5*0.25, 1.0,
				   0.5+0.5*0.25, 0.0, 0.5+1.5*0.25, 1.0);
    break;
  case 0x2550:			/* BOX DRAWINGS DOUBLE HORIZONTAL */
    glyph = font_render_line_char (c, fi, forecol, backcol,
				   line_pos_ab, line_weight_double,
				   line_pos_none, line_weight_double);
    break;
  case 0x2551:			/* BOX DRAWINGS DOUBLE VERTICAL */
    glyph = font_render_line_char (c, fi, forecol, backcol,
				   line_pos_none, line_weight_double,
				   line_pos_ab, line_weight_double);
    break;
  case 0x2552:	       /* BOX DRAWINGS DOWN SINGLE AND RIGHT DOUBLE */
    glyph = font_render_line_char (c, fi, forecol, backcol,
				   line_pos_b, line_weight_double,
				   line_pos_b, line_weight_light);
    break;
  case 0x2553:	       /* BOX DRAWINGS DOWN DOUBLE AND RIGHT SINGLE */
    glyph = font_render_line_char (c, fi, forecol, backcol,
				   line_pos_b, line_weight_light,
				   line_pos_b, line_weight_double);
    break;
  case 0x2554:		      /* BOX DRAWINGS DOUBLE DOWN AND RIGHT */
    glyph = font_render_line_char (c, fi, forecol, backcol,
				   line_pos_b, line_weight_double,
				   line_pos_b, line_weight_double);
    break;
  case 0x2555:		/* BOX DRAWINGS DOWN SINGLE AND LEFT DOUBLE */
    glyph = font_render_line_char (c, fi, forecol, backcol,
				   line_pos_a, line_weight_double,
				   line_pos_b, line_weight_light);
    break;
  case 0x2556:		/* BOX DRAWINGS DOWN DOUBLE AND LEFT SINGLE */
    glyph = font_render_line_char (c, fi, forecol, backcol,
				   line_pos_a, line_weight_light,
				   line_pos_b, line_weight_double);
    break;
  case 0x2557:		       /* BOX DRAWINGS DOUBLE DOWN AND LEFT */
    glyph = font_render_line_char (c, fi, forecol, backcol,
				   line_pos_a, line_weight_double,
				   line_pos_b, line_weight_double);
    break;
  case 0x2558:		 /* BOX DRAWINGS UP SINGLE AND RIGHT DOUBLE */
    glyph = font_render_line_char (c, fi, forecol, backcol,
				   line_pos_b, line_weight_double,
				   line_pos_a, line_weight_light);
    break;
  case 0x2559:		 /* BOX DRAWINGS UP DOUBLE AND RIGHT SINGLE */
    glyph = font_render_line_char (c, fi, forecol, backcol,
				   line_pos_b, line_weight_light,
				   line_pos_a, line_weight_double);
    break;
  case 0x255A:			/* BOX DRAWINGS DOUBLE UP AND RIGHT */
    glyph = font_render_line_char (c, fi, forecol, backcol,
				   line_pos_b, line_weight_double,
				   line_pos_a, line_weight_double);
    break;
  case 0x255B:		  /* BOX DRAWINGS UP SINGLE AND LEFT DOUBLE */
    glyph = font_render_line_char (c, fi, forecol, backcol,
				   line_pos_a, line_weight_double,
				   line_pos_a, line_weight_light);
    break;
  case 0x255C:		  /* BOX DRAWINGS UP DOUBLE AND LEFT SINGLE */
    glyph = font_render_line_char (c, fi, forecol, backcol,
				   line_pos_a, line_weight_light,
				   line_pos_a, line_weight_double);
    break;
  case 0x255D:			/* BOX DRAWINGS DOUBLE UP AND LEFT */
    glyph = font_render_line_char (c, fi, forecol, backcol,
				   line_pos_a, line_weight_double,
				   line_pos_a, line_weight_double);
    break;
  case 0x255E:	   /* BOX DRAWINGS VERTICAL SINGLE AND RIGHT DOUBLE */
    glyph = font_render_line_char (c, fi, forecol, backcol,
				   line_pos_b, line_weight_double,
				   line_pos_ab, line_weight_light);
    break;
  case 0x255F:	   /* BOX DRAWINGS VERTICAL DOUBLE AND RIGHT SINGLE */
    glyph = font_render_line_char (c, fi, forecol, backcol,
				   line_pos_b, line_weight_light,
				   line_pos_ab, line_weight_double);
    break;
  case 0x2560:		  /* BOX DRAWINGS DOUBLE VERTICAL AND RIGHT */
    glyph = font_render_line_char (c, fi, forecol, backcol,
				   line_pos_b, line_weight_double,
				   line_pos_ab, line_weight_double);
    break;
  case 0x2561:		  /* BOX DRAWINGS VERTICAL SINGLE AND LEFT DOUBLE */
    glyph = font_render_line_char (c, fi, forecol, backcol,
				   line_pos_a, line_weight_double,
				   line_pos_ab, line_weight_light);
    break;
  case 0x2562:		  /* BOX DRAWINGS VERTICAL DOUBLE AND LEFT SINGLE */
    glyph = font_render_line_char (c, fi, forecol, backcol,
				   line_pos_a, line_weight_light,
				   line_pos_ab, line_weight_double);
    break;
  case 0x2563:		  /* BOX DRAWINGS DOUBLE VERTICAL AND LEFT */
    glyph = font_render_line_char (c, fi, forecol, backcol,
				   line_pos_a, line_weight_double,
				   line_pos_ab, line_weight_double);
    break;
  case 0x2564:	  /* BOX DRAWINGS DOWN SINGLE AND HORIZONTAL DOUBLE */
    glyph = font_render_line_char (c, fi, forecol, backcol,
				   line_pos_ab, line_weight_double,
				   line_pos_b, line_weight_light);
    break;
  case 0x2565:	  /* BOX DRAWINGS DOWN DOUBLE AND HORIZONTAL SINGLE */
    glyph = font_render_line_char (c, fi, forecol, backcol,
				   line_pos_ab, line_weight_light,
				   line_pos_b, line_weight_double);
    break;
  case 0x2566:		 /* BOX DRAWINGS DOUBLE DOWN AND HORIZONTAL */
    glyph = font_render_line_char (c, fi, forecol, backcol,
				   line_pos_ab, line_weight_double,
				   line_pos_b, line_weight_double);
    break;
  case 0x2567:	    /* BOX DRAWINGS UP SINGLE AND HORIZONTAL DOUBLE */
    glyph = font_render_line_char (c, fi, forecol, backcol,
				   line_pos_ab, line_weight_double,
				   line_pos_a, line_weight_light);
    break;
  case 0x2568:	    /* BOX DRAWINGS UP DOUBLE AND HORIZONTAL SINGLE */
    glyph = font_render_line_char (c, fi, forecol, backcol,
				   line_pos_ab, line_weight_light,
				   line_pos_a, line_weight_double);
    break;
  case 0x2569:	    /* BOX DRAWINGS UP DOUBLE AND HORIZONTAL */
    glyph = font_render_line_char (c, fi, forecol, backcol,
				   line_pos_ab, line_weight_double,
				   line_pos_a, line_weight_double);
    break;
  case 0x256A: /* BOX DRAWINGS VERTICAL SINGLE AND HORIZONTAL DOUBLE */
    glyph = font_render_line_char (c, fi, forecol, backcol,
				   line_pos_ab, line_weight_double,
				   line_pos_ab, line_weight_light);
    break;
  case 0x256B: /* BOX DRAWINGS VERTICAL DOUBLE AND HORIZONTAL SINGLE */
    glyph = font_render_line_char (c, fi, forecol, backcol,
				   line_pos_ab, line_weight_light,
				   line_pos_ab, line_weight_double);
    break;
  case 0x256C:	     /* BOX DRAWINGS DOUBLE VERTICAL AND HORIZONTAL */
    glyph = font_render_line_char (c, fi, forecol, backcol,
				   line_pos_ab, line_weight_double,
				   line_pos_ab, line_weight_double);
    break;
  case 0x2574:			/* BOX DRAWINGS LIGHT LEFT */
    glyph = font_render_line_char (c, fi, forecol, backcol,
				   line_pos_a, line_weight_light,
				   line_pos_none, line_weight_light);
    break;
  case 0x2575:			/* BOX DRAWINGS LIGHT UP */
    glyph = font_render_line_char (c, fi, forecol, backcol,
				   line_pos_none, line_weight_light,
				   line_pos_a, line_weight_light);
    break;
  case 0x2576:			/* BOX DRAWINGS LIGHT RIGHT */
    glyph = font_render_line_char (c, fi, forecol, backcol,
				   line_pos_b, line_weight_light,
				   line_pos_none, line_weight_light);
    break;
  case 0x2577:			/* BOX DRAWINGS LIGHT DOWN */
    glyph = font_render_line_char (c, fi, forecol, backcol,
				   line_pos_none, line_weight_light,
				   line_pos_b, line_weight_light);
    break;
  case 0x2578:			/* BOX DRAWINGS HEAVY LEFT */
    glyph = font_render_line_char (c, fi, forecol, backcol,
				   line_pos_a, line_weight_heavy,
				   line_pos_none, line_weight_heavy);
    break;
  case 0x2579:			/* BOX DRAWINGS HEAVY UP */
    glyph = font_render_line_char (c, fi, forecol, backcol,
				   line_pos_none, line_weight_heavy,
				   line_pos_a, line_weight_heavy);
    break;
  case 0x257A:			/* BOX DRAWINGS HEAVY RIGHT */
    glyph = font_render_line_char (c, fi, forecol, backcol,
				   line_pos_b, line_weight_heavy,
				   line_pos_none, line_weight_heavy);
    break;
  case 0x257B:			/* BOX DRAWINGS HEAVY DOWN */
    glyph = font_render_line_char (c, fi, forecol, backcol,
				   line_pos_none, line_weight_heavy,
				   line_pos_b, line_weight_heavy);
    break;
  case 0x2580:			/* UPPER HALF BLOCK */
    glyph = font_render_box_char (c, fi, forecol, backcol, 0, 0, 1, 0.5);
    break;
  case 0x2581:			/* LOWER ONE EIGHTH BLOCK */
    glyph = font_render_box_char (c, fi, forecol, backcol, 0, 7/8.0, 1, 1);
    break;
  case 0x2582:			/* LOWER ONE QUARTER BLOCK */
    glyph = font_render_box_char (c, fi, forecol, backcol, 0, 3/4.0, 1, 1);
    break;
  case 0x2583:			/* LOWER THREE EIGHTHS BLOCK */
    glyph = font_render_box_char (c, fi, forecol, backcol, 0, 5/8.0, 1, 1);
    break;
  case 0x2584:			/* LOWER HALF BLOCK */
    glyph = font_render_box_char (c, fi, forecol, backcol, 0, 0.5, 1, 1);
    break;
  case 0x2585:			/* LOWER FIVE EIGHTHS BLOCK */
    glyph = font_render_box_char (c, fi, forecol, backcol, 0, 3/8.0, 1, 1);
    break;
  case 0x2586:			/* LOWER THREE QUARTERS BLOCK */
    glyph = font_render_box_char (c, fi, forecol, backcol, 0, 1/4.0, 1, 1);
    break;
  case 0x2587:			/* LOWER SEVEN EIGHTHS BLOCK */
    glyph = font_render_box_char (c, fi, forecol, backcol, 0, 1/8.0, 1, 1);
    break;
  case 0x2588:			/* FULL BLOCK */
    glyph = font_render_box_char (c, fi, forecol, backcol, 0, 0, 1, 1);
    break;
  case 0x2589:			/* LEFT SEVEN EIGHTHS BLOCK */
    glyph = font_render_box_char (c, fi, forecol, backcol, 0, 0, 7/8.0, 1);
    break;
  case 0x258A:			/* LEFT THREE QUARTERS BLOCK */
    glyph = font_render_box_char (c, fi, forecol, backcol, 0, 0, 3/4.0, 1);
    break;
  case 0x258B:			/* LEFT FIVE EIGHTHS BLOCK */
    glyph = font_render_box_char (c, fi, forecol, backcol, 0, 0, 5/8.0, 1);
    break;
  case 0x258C:			/* LEFT HALF BLOCK */
    glyph = font_render_box_char (c, fi, forecol, backcol, 0, 0, 0.5, 1);
    break;
  case 0x258D:			/* LEFT THREE EIGHTHS BLOCK */
    glyph = font_render_box_char (c, fi, forecol, backcol, 0, 0, 3/8.0, 1);
    break;
  case 0x258E:			/* LEFT ONE QUARTER BLOCK */
    glyph = font_render_box_char (c, fi, forecol, backcol, 0, 0, 1/4.0, 1);
    break;
  case 0x258F:			/* LEFT ONE EIGHTH BLOCK */
    glyph = font_render_box_char (c, fi, forecol, backcol, 0, 0, 1/8.0, 1);
    break;
  case 0x2590:			/* RIGHT HALF BLOCK */
    glyph = font_render_box_char (c, fi, forecol, backcol, 0.5, 0, 1, 1);
    break;
  case 0x2591:			/* LIGHT SHADE */
    glyph = font_render_shade (c, fi, forecol, backcol, 0.25);
    break;
  case 0x2592:			/* MEDIUM SHADE */
    glyph = font_render_shade (c, fi, forecol, backcol, 0.5);
    break;
  case 0x2593:			/* DARK SHADE */
    glyph = font_render_shade (c, fi, forecol, backcol, 0.75);
    break;
  case 0x2594:			/* UPPER ONE EIGHTH BLOCK */
    glyph = font_render_box_char (c, fi, forecol, backcol, 0, 0, 1, 0.125);
    break;
  case 0x2595:			/* RIGHT ONE EIGHTH BLOCK */
    glyph = font_render_box_char (c, fi, forecol, backcol, 1-0.125, 0, 1, 1);
    break;
  default:
    glyph = font_render_ft_glyph (c, fi, forecol, backcol);
    break;
  }
  return  glyph;
}

static SDL_Surface *
font_create_combined_ft_glyph (SDL_Surface *orig,
			       Uint16 c, struct font_info *fi,
			       SDL_Color *forecol, SDL_Color *backcol)
/* Superimpose ORIG with character C from the loaded font.  Return the
 * result as a newly allocated SDL surface, or NULL if the glyph is
 * not available.  */
{
  SDL_Surface *res;
  int  rc, x_offs, y_offs;
  
  rc = FT_Load_Char (face, c, FT_LOAD_RENDER);
  if (rc)  return  NULL;

  x_offs = face->glyph->bitmap_left;
  x_offs += 0.5*(fi->cell_width - face->glyph->metrics.horiAdvance/64.0);
  y_offs = fi->cell_height - (face->glyph->bitmap_top + fi->descent);

  res = SDL_DisplayFormat (orig);
  overlay (res, face->glyph->bitmap.buffer,
	   face->glyph->bitmap.width, face->glyph->bitmap.rows,
	   face->glyph->bitmap.pitch, x_offs, y_offs, forecol);
  return  res;
}

static SDL_Surface *
create_acs_glyph (Uint16 c, struct font_info *fi,
		  SDL_Color *forecol, SDL_Color *backcol)
/* Create a SDL surface containing a glyph for the ACS character c.
 * Return NULL if no glyph is available.  Most characters are
 * translated into their Unicode equivalents, the remaining glyphs are
 * rendered manually.  */
{
  int  w = fi->cell_width;
  int  h = fi->cell_height;
  unsigned char  data [w*h];
  double  dx, dy;
  int  i, j;

  switch (c) {
  case '*':			/* ACS_BULLET   */
    return  font_create_unicode_glyph (0x2022, fi, forecol, backcol);
  case '+':			/* ACS_RARROW   */
    return  font_create_unicode_glyph (0x2192, fi, forecol, backcol);
  case ',':			/* ACS_LARROW   */
    return  font_create_unicode_glyph (0x2190, fi, forecol, backcol);
  case '-':			/* ACS_UARROW   */
    return  font_create_unicode_glyph (0x2191, fi, forecol, backcol);
  case '.':			/* ACS_DARROW   */
    return  font_create_unicode_glyph (0x2193, fi, forecol, backcol);
  case '0':			/* ACS_BLOCK    */
    return  font_create_unicode_glyph (0x2588, fi, forecol, backcol);
  case '1':
    return  font_create_unicode_glyph (0x2593, fi, forecol, backcol);
  case '2':
    return  font_create_unicode_glyph (0x2592, fi, forecol, backcol);
  case '3':
    return  font_create_unicode_glyph (0x2591, fi, forecol, backcol);
  case 'C':			/* nethack: dagger */
    return  font_create_unicode_glyph (0x2020, fi, forecol, backcol);
  case 'J':
    return  font_create_unicode_glyph (0x2518, fi, forecol, backcol);
  case 'K':
    return  font_create_unicode_glyph (0x2510, fi, forecol, backcol);
  case 'L':
    return  font_create_unicode_glyph (0x250C, fi, forecol, backcol);
  case 'M':
    return  font_create_unicode_glyph (0x2514, fi, forecol, backcol);
  case 'N':
    return  font_create_unicode_glyph (0x253C, fi, forecol, backcol);
  case 'Q':
    return  font_create_unicode_glyph (0x2500, fi, forecol, backcol);
  case 'T':
    return  font_create_unicode_glyph (0x251C, fi, forecol, backcol);
  case 'U':
    return  font_create_unicode_glyph (0x2524, fi, forecol, backcol);
  case 'V':
    return  font_create_unicode_glyph (0x2534, fi, forecol, backcol);
  case 'W':
    return  font_create_unicode_glyph (0x252C, fi, forecol, backcol);
  case 'X':
    return  font_create_unicode_glyph (0x2502, fi, forecol, backcol);
  case '[':			/* nethack: integral, top half */
    return  font_create_unicode_glyph (0x2320, fi, forecol, backcol);
  case ']':
    return  font_create_unicode_glyph (0x2321, fi, forecol, backcol);
  case 'f':			/* ACS_DEGREE   */
    return  font_create_unicode_glyph (0x00B0, fi, forecol, backcol);
  case 'g':			/* ACS_PLMINUS  */
    return  font_create_unicode_glyph (0x00B1, fi, forecol, backcol);
  case 'j':			/* ACS_LRCORNER */
    return  font_create_unicode_glyph (0x251B, fi, forecol, backcol);
  case 'k':			/* ACS_URCORNER */
    return  font_create_unicode_glyph (0x2513, fi, forecol, backcol);
  case 'l':			/* ACS_ULCORNER */
    return  font_create_unicode_glyph (0x250F, fi, forecol, backcol);
  case 'm':			/* ACS_LLCORNER */
    return  font_create_unicode_glyph (0x2517, fi, forecol, backcol);
  case 'n':			/* ACS_PLUS     */
    return  font_create_unicode_glyph (0x254B, fi, forecol, backcol);
  case 'q':			/* ACS_HLINE    */
    return  font_create_unicode_glyph (0x2501, fi, forecol, backcol);
  case 't':			/* ACS_LTEE     */
    return  font_create_unicode_glyph (0x2523, fi, forecol, backcol);
  case 'u':			/* ACS_RTEE     */
    return  font_create_unicode_glyph (0x252B, fi, forecol, backcol);
  case 'v':			/* ACS_BTEE     */
    return  font_create_unicode_glyph (0x253B, fi, forecol, backcol);
  case 'w':			/* ACS_TTEE     */
    return  font_create_unicode_glyph (0x2533, fi, forecol, backcol);
  case 'x':			/* ACS_VLINE    */
    return  font_create_unicode_glyph (0x2503, fi, forecol, backcol);
  case 'y':			/* ACS_LEQUAL   */
    return  font_create_unicode_glyph (0x2264, fi, forecol, backcol);
  case 'z':			/* ACS_GEQUAL   */
    return  font_create_unicode_glyph (0x2265, fi, forecol, backcol);
  case '{':			/* ACS_PI       */
    return  font_create_unicode_glyph (0x03C0, fi, forecol, backcol);
  case '|':			/* ACS_NEQUAL   */
    return  font_create_unicode_glyph (0x2260, fi, forecol, backcol);
  case '}':			/* ACS_STERLING */
    return  font_create_unicode_glyph (0x00A3, fi, forecol, backcol);
  }

  for (i=0; i<w*h; ++i)  data[i] = 0;
  dx = 0.125;
  dy = 0.5*dx;

  switch (c) {
  case '`':			/* ACS_DIAMOND  */
    for (i=0; i<w; ++i) {
      for (j=0; j<h; ++j) {
	double d;
	d = fabs(j-0.5*h)/h + fabs(i-0.5*w)/w;
	d = 4-d*10;
	if (d<0) d=0;
	if (d>0.99) d=0.99;
	data[i+w*j] = 256*d;
      }
    }
    break;
  case 'a':			/* nethack: solid block */
    for (i=0; i<w; ++i) {
      for (j=0; j<h; ++j) {
	data[i+w*j] = (i+j)%2 ? 255 : 0;
      }
    }
    break;
  case 'b':			/* ACS_CKBOARD  */
    {
      int  nx, ny;
      nx = 2*(int)(w/6);
      if (nx<4)  nx=4;
      ny = 2*nx;
      for (i=0; i<nx; ++i) {
	for (j=0; j<ny; ++j) {
	  if ((i+j) % 2) continue;
	  draw_rectangle (fi, data,
			  i/(double)nx, j/(double)ny,
			  (i+1)/(double)nx, (j+1)/(double)ny,
			  0);
	}
      }
    }
    break;
  case 'h':			/* ACS_BOARD    */
    {
      int  nx, ny;
      nx = 2*(int)(w/12);
      if (nx<2)  nx=2;
      ny = 2*nx;
      for (i=0; i<nx; ++i) {
	for (j=0; j<ny; ++j) {
	  if ((i+j) % 2) continue;
	  draw_rectangle (fi, data,
			  i/(double)nx, j/(double)ny,
			  (i+1)/(double)nx, (j+1)/(double)ny,
			  0);
	}
      }
    }
    break;
  case 'i':			/* ACS_LANTERN  */
    draw_rectangle (fi, data, 0.15, 0.15, 0.85, 0.85, 0);
    break;
  case 'o':			/* ACS_S1       */
    draw_rectangle (fi, data, 0, 0/9.0, 1, 1/9.0, 0);
    break;
  case 'p':			/* ACS_S3       */
    draw_rectangle (fi, data, 0, 2/9.0, 1, 3/9.0, 0);
    break;
  case 'r':			/* ACS_S7       */
    draw_rectangle (fi, data, 0, 6/9.0, 1, 7/9.0, 0);
    break;
  case 's':			/* ACS_S9       */
    draw_rectangle (fi, data, 0, 8/9.0, 1, 9/9.0, 0);
    break;
  case '~':			/* nethack: centered dot */
    draw_rectangle (fi, data,
		    0.5-1.0/16, 0.5+1.0/16, 0.5-1.0/16, 0.5+1.0/16, 0);
    break;
  default:
    return  NULL;
  }

  return  convert_to_surface (data, w, h, w, 0, 0, fi, forecol, backcol);
}

SDL_Surface *
font_get_glyph (struct font_info *fi, struct charcell cc)
/* Get the SDL_Surface containing the glyph for character CC.
 * This function tries to get the glyph from the cache first.
 * It renders new glyphs as necessary.
 *
 * The returned SDL_Surface is owned by the font cache and may
 * not be freed by the caller.  */
{
  struct hash_entry *ptr;
  int  present;

  present = hash_get_entry (cc, &ptr);
  if (! present) {
    SDL_Color *forecol, *backcol;

    color_decode (cc, &forecol, &backcol);
    if (cc.attr & JV_ATTR_COMBINED) {
      Uint16  c1, c2;

      if (charcell_disband (cc, &c1, &c2)) {
	ptr->glyph = font_create_unicode_glyph (c1, fi, forecol, backcol);
	if (ptr->glyph) {
	  ptr->glyph = font_create_combined_ft_glyph (ptr->glyph, c2, fi,
						      forecol, backcol);
	}
      } else {
	fatal ("invalid combined glyph %04X", cc.c);
      }
    } else if (cc.attr & JV_ATTR_ACS) {
      ptr->glyph = create_acs_glyph (cc.c, fi, forecol, backcol);
    } else {
      ptr->glyph = font_create_unicode_glyph (cc.c, fi, forecol, backcol);
    }
    
    if (! ptr->glyph) {
      if (cc.attr & JV_ATTR_COMBINED) {
	Uint16  c1, c2;
	charcell_disband (cc, &c1, &c2);
	/* TODO: handle glyphs made up of more than two characters */
	warning ("cannot render combined glyph U+%04X+U+%04X", c1, c2);
      } else if (cc.attr & JV_ATTR_ACS) {
	warning ("cannot render glyph for ACS character 0x%04X", cc.c);
      } else {
	warning ("cannot render glyph for UCS2 character 0x%04X", cc.c);
      }
      ptr->glyph = font_create_unicode_glyph (0xFFFD, fi,
					      forecol, backcol);
      if (! ptr->glyph)  fatal ("cannot render replacement character");
    }

    if (cc.attr & JV_ATTR_UNDERLINE)  underline (ptr->glyph, forecol, fi);
  }
  return  ptr->glyph;
}
