/* jvterm.c - the world's brightest terminal emulator
 *
 * Copyright (C) 2003, 2005  Jochen Voss
 *
 * This program is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 2 of the License, or
 * (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 51 Franklin St, Fifth Floor, Boston, MA  02110-1301 USA
 *
 * $Id: jvterm.c 6725 2005-11-09 23:15:58Z voss $
 */

#ifdef HAVE_CONFIG_H
#  include <config.h>
#endif

#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <unistd.h>
#include <sys/select.h>
#include <sys/wait.h>
#include <signal.h>
#include <errno.h>

#include <locale.h>
#include <nl_types.h>
#include <langinfo.h>

#include <SDL.h>
#include <SDL_thread.h>

#include "base.h"
#include "backend.h"
#include "frontend.h"
#include "ui.h"


static char *parent_CHARSET = NULL;
static char *child_CHARSET = NULL;
static char *child_LC_CTYPE = NULL;

static char *default_shell = NULL;

static  int  nchildren;
static  int  control_flag = 0;
int  shutdown_flag = 0;

static  int  magic_modifier = (KMOD_ALT|KMOD_META);

static  struct client *dddlient;

/**********************************************************************
 * user events
 */

#define EV_BLINK 59702
#define EV_FLASH 54484

/**********************************************************************
 * emit message to stdout and stderr
 */

struct stdlog {
  struct converter *conv;
  int  verbose_flag;
  FILE *fstd, *ferr, *f;
  int  handle;
};

static struct stdlog *logcon = NULL;
static struct stdlog *logfile = NULL;

static ssize_t
std_emit (const void *data, size_t size, void *client_data)
{
  struct stdlog *l = client_data;
  return  fwrite (data, 1, size, l->f);
}

static void
std_write_line (enum loglevel level, const char *message,
		void *client_data)
{
  struct stdlog *l = client_data;
  const char *prefix = "??? ";

  l->f = l->ferr;
  switch (level) {
  case log_MESSAGE:
    if (! l->verbose_flag)  return;
    prefix = "";
    l->f = l->fstd;
    break;
  case log_WARNING:
    if (! l->verbose_flag)  return;
    prefix = "warning: ";
    break;
  case log_ERROR:
    prefix = "error: ";
    break;
  }
  converter_write (l->conv, prefix, strlen(prefix));
  converter_write (l->conv, message, strlen(message));
  converter_write (l->conv, "\n", 1);
  converter_reset (l->conv);
  fflush (l->f);
}

static void
open_std_logs (int verbose_flag, const char *logfile_name)
{
  logcon = xnew (struct stdlog);
  logcon->conv = open_converter ("stdout/stderr", "UTF-8", parent_CHARSET,
				 "", 0, std_emit, logcon);
  logcon->verbose_flag = verbose_flag;
  logcon->fstd = stdout;
  logcon->ferr = stderr;
  logcon->f = stderr;
  logcon->handle = log_add_client (std_write_line, logcon);

  if (logfile_name) {
    FILE *f;

    f = fopen (logfile_name, "w");
    if (f == NULL) {
      error ("cannot open logfile: %m");
      return;
    }
    
    logfile = xnew (struct stdlog);
    logfile->verbose_flag = 1;
    logfile->fstd = f;
    logfile->ferr = f;
    logfile->f = f;
    logfile->conv = open_converter ("logfile", "UTF-8", parent_CHARSET,
				    "", 0, std_emit, logfile);
    logfile->handle = log_add_client (std_write_line, logfile);
  } else {
    logfile = NULL;
  }
}

static void
close_std_logs (void)
{
  if (logfile) {
    log_remove_client (logfile->handle);
    close_converter (logfile->conv);
    fclose (logfile->f);
    xfree (logfile);
    logfile = NULL;
  }

  log_remove_client (logcon->handle);
  close_converter (logcon->conv);
  xfree (logcon);
  logcon = NULL;
}

/**********************************************************************
 * register file descriptors where we wait for input
 */

struct listener_info {
  int fd;
  void *data;
};
static  struct listener_info *listeners;
static  int  listeners_allocated = 0;
static  int  listeners_used = 0;

static void
listener_add (int fd, void *data)
{
  if (listeners_used >= listeners_allocated) {
    listeners_allocated += 8;
    listeners = xrealloc (listeners,
			  listeners_allocated*sizeof(struct listener_info));
  }
  listeners[listeners_used].fd = fd;
  listeners[listeners_used].data = data;
  ++listeners_used;
}

static void
listener_remove (int fd)
{
  int  i;
  
  for (i=0; i<=listeners_used; ++i) {
    if (listeners[i].fd != fd)  continue;
    memmove (listeners+i, listeners+i+1,
	     (listeners_used-(i+1))*sizeof(struct listener_info));
    --listeners_used;
    break;
  }
}

/**********************************************************************
 * a shell client
 */

#include <unistd.h>
#include <fcntl.h>
#include <ctype.h>

struct client *vt[8];
int  current_vt = -1;

static ssize_t
do_real_send_converted (const void *buffer, size_t size, void *data)
{
  struct shell_client_data *scd = data;
  return  write (scd->fd, buffer, size);
}

     
static int
set_nonblock_flag (int desc, int value)
/* from GNU the libc manual */
/* Set the `O_NONBLOCK' flag of DESC if VALUE is nonzero,
   or clear the flag if VALUE is 0.
   Return 0 on success, or -1 on error with `errno' set. */
{
  int oldflags = fcntl (desc, F_GETFL, 0);
  /* If reading the flags failed, return error indication now. */
  if (oldflags == -1)
    return -1;
  /* Set just the flag we want to set. */
  if (value != 0)
    oldflags |= O_NONBLOCK;
  else
    oldflags &= ~O_NONBLOCK;
  /* Store modified flag word in the descriptor. */
  return fcntl (desc, F_SETFL, oldflags);
}

static void
fix_key_quartz (SDL_KeyboardEvent *key)
{
  SDL_keysym *keysym = &(key->keysym);
  if (keysym->mod&KMOD_SHIFT && keysym->unicode==',') {
    keysym->mod &= ~KMOD_SHIFT;
    keysym->unicode='<';
  }
  if (keysym->mod&KMOD_SHIFT && keysym->unicode=='.') {
    keysym->mod &= ~KMOD_SHIFT;
    keysym->unicode='>';
  }
}

static const struct {
  const char *driver;
  void (*fix_key) (SDL_KeyboardEvent *key);
} fix_key_functions[] = {
  { "Quartz", fix_key_quartz },
  { NULL, NULL }
};

static void *
shell_init (struct buffer *buf, void *init_data)
{
  char  buffer [80];
  struct shell_client_init *sci = init_data;
  struct shell_client_data *scd;
  struct shell_info si;
  int  i, rc;
  
  scd = xnew (struct shell_client_data);
  scd->child_charset = xstrdup (sci->child_charset);
  sprintf (buffer, "child->console%d", sci->vt+1);
  scd->con = console_open (buf, scd, buffer);

  scd->fix_key = NULL;
  SDL_VideoDriverName (buffer, 80);
  for (i=0; fix_key_functions[i].driver; ++i) {
    if (strcmp (fix_key_functions[i].driver, buffer) == 0) {
      scd->fix_key = fix_key_functions[i].fix_key;
      message ("using %.80s driver key fixes", buffer);
    }
  }

  /* prepare to convert the shell's keyboard input */
  sprintf (buffer, "console%d->child", sci->vt+1);
  scd->conv_input = open_converter (buffer,
				    "UCS-2", sci->child_charset,
				    "", 0, do_real_send_converted, scd);
  
  scd->fd = -1;			/* TODO: is this necessary? */

  si.lc_ctype = child_LC_CTYPE;
  si.shell = default_shell;
  si.meta_mode = sci->meta_mode;
  si.cols = screen_get_cols();
  si.rows = screen_get_rows();
  scd->fd = spawn_shell (&si, &scd->pid);
  rc = set_nonblock_flag (scd->fd, 1);
  if (rc != 0) {
    error ("cannot set non-blocking mode: %m");
  }

  scd->meta_mode = sci->meta_mode;
  scd->vt = sci->vt;

  ++nchildren;

  /* prepare to process the shell output */
  listener_add (scd->fd, scd);

  /* Note to the confused reader: The flow of control is a little bit
   * non-obvious here.  Keyboard input is fed to the shell child
   * process via the 'conv_input' converter, allocated above.  The
   * shell output is sent to the console for display via the
   * 'console_copy' call in 'listen_input'.  */
  
  return  scd;
}

static void
shell_exit (void *data)
{
  struct shell_client_data *scd = data;
  pid_t  rc;
  int  shell_rc;

 retry_wait:
  rc = waitpid (scd->pid, &shell_rc, WNOHANG);
  if (rc<0) {
    if (errno == EINTR)  goto retry_wait;
    if (errno != ECHILD)  error ("shell_exit: waitpid failed: %m");
  } else if (rc != scd->pid) {
    kill (scd->pid, SIGTERM);
    usleep (200000);
    kill (scd->pid, SIGKILL);
    rc = waitpid (scd->pid, &shell_rc, 0);
  }
  if (WIFEXITED (shell_rc)) {
    message ("vt%02d: child exited with exit status %d",
	     scd->vt, WEXITSTATUS(shell_rc));
  } else if (WIFSIGNALED (shell_rc)) {
    message ("vt%02d: child terminated by signal %d",
	     scd->vt, WTERMSIG(shell_rc));
  } else if (WIFSTOPPED (shell_rc)) {
    message ("vt%02d: child stopped by signal %d",
	     scd->vt, WSTOPSIG(shell_rc));
  } else {
    warning ("vt%02d: child is dead rc=%d", scd->vt, shell_rc);
  }
  
  listener_remove (scd->fd);
  
  close_converter (scd->conv_input);
  console_close (scd->con);
  xfree (scd->child_charset);
  xfree (scd);

  --nchildren;
}

void
shell_send_char (struct shell_client_data *scd, Uint16 c)
/* Feed a char into the shell client's input stream.
 * TODO: use client_send_char */
{
  converter_write (scd->conv_input, &c, 2);
}

static void
shell_send_char_meta (struct shell_client_data *scd, Uint16 c, int meta)
{
  if (meta) {
    if (scd->meta_mode) {
      shell_send_char (scd, c|128);
    } else {
      shell_send_char (scd, 27);
      shell_send_char (scd, c);
    }
  } else {
    shell_send_char (scd, c);
  }
}

#define META_MASK (KMOD_ALT|KMOD_META)

static void
shell_send_key (SDL_KeyboardEvent *key, void *data)
{
  struct shell_client_data *scd = data;
  SDL_keysym *keysym = &(key->keysym);
  
  switch (keysym->sym) {
  case SDLK_BACKSPACE:
    shell_send_char_meta (scd, 127, keysym->mod&META_MASK);
    break;
  case SDLK_ESCAPE:
    shell_send_char_meta (scd, 27, keysym->mod&META_MASK);
    break;
	
#include "keys.h"
	
  default:
    if (scd->fix_key)  scd->fix_key (key);
    if (keysym->unicode) {
      shell_send_char_meta (scd, keysym->unicode, keysym->mod&META_MASK);
    } else if (keysym->mod&KMOD_CTRL
	       && keysym->sym == 32) {
      shell_send_char (scd, 0);	/* CTRL-Space */
    }
    break;
  }
}

ssize_t
shell_send_string (Uint16 *s, size_t len, void *data)
{
  struct shell_client_data *scd = data;
  converter_write (scd->conv_input, s, len);
  return  len;
}

static void *
shell_get_pid (void *cdata, void *data)
{
  struct shell_client_data *scd = cdata;
  return  &(scd->pid);
}

struct client_info shell_client = {
  init: shell_init,
  exit: shell_exit,
  activate: NULL,
  deactivate: NULL,
  send_key: shell_send_key,
  send_string: shell_send_string,
};

void
select_vt (int n, int create)
{
  if (n<0) {
    for (n=0; n<8; ++n) {
      if (vt[n])  break;
    }
  }
  if (n<8 && create && ! vt[n]) {
    struct shell_client_init sci = {
      child_charset: child_CHARSET,
      meta_mode : (create == 2), /* TODO: hack alert */
      vt: n
    };
    vt[n] = new_client (&shell_client, &sci, 1000);
  }
  if (n<8) {
    current_vt = n;
    control_flag = 0;
  }
  if (n==11) {
    client_activate (messages);
  } else if (n<8 && vt[n]) {
    client_activate (vt[n]);
  } else {
    client_activate (dddlient);
  }
}

/**********************************************************************
 * visual bell
 */

static  SDL_TimerID  flash_timer;
static  int  flash_count = 0;

static Uint32
flash_cb (Uint32 interval, void *param)
{
  SDL_Event event;

  if (flash_count <= 0) {
    SDL_RemoveTimer (flash_timer);
  } else {
    flash_count -= 1;
    
    event.type = SDL_USEREVENT;
    event.user.code = EV_FLASH;
    event.user.data1 = 0;
    event.user.data2 = 0;
    SDL_PushEvent (&event);
  }

  return  interval;
}

static void
start_flash (int n)
{
  flash_count = n;
  flash_timer = SDL_AddTimer (50, flash_cb, NULL);
}

/**********************************************************************
 * main loop
 */

static  Uint32  ctrl_history, shift_history;

static  SDL_TimerID  blink_timer;

static void
quit_sdl (void)
/* Shutdown all subsystems.  This is called via 'atexit' */
{
  SDL_Quit ();
}

static Uint32
blink_cb (Uint32 interval, void *param)
{
  SDL_Event event;

  event.type = SDL_USEREVENT;
  event.user.code = EV_BLINK;
  event.user.data1 = 0;
  event.user.data2 = 0;
  SDL_PushEvent (&event);

  return  interval;
}

static void
handle_event (SDL_Event *event)
{
  int  key_done = 0;
  
  switch (event->type) {
  case SDL_KEYDOWN:
    switch (event->key.keysym.sym) {
    case SDLK_LCTRL:
    case SDLK_RCTRL:
      {
	Uint32  now = SDL_GetTicks ();

	if (now-ctrl_history < 400) {
	  ctrl_history = 0;
	  client_activate (dddlient);
	  key_done = 1;
	} else {
	  ctrl_history = now;
	}
      }
      break;
    case SDLK_LSHIFT:
    case SDLK_RSHIFT:
      {
	Uint32  now = SDL_GetTicks ();

	if (now-shift_history < 400) {
	  shift_history = 0;
	  client_start_select ();
	  key_done = 1;
	} else {
	  shift_history = now;
	}
      }
      break;
    case SDLK_LEFT:
      if (event->key.keysym.mod&magic_modifier) {
	if (current_vt > 0)  select_vt (current_vt-1, 0);
	key_done = 1;
      }
      break;
    case SDLK_RIGHT:
      if (event->key.keysym.mod&magic_modifier) {
	if (current_vt < 7)  select_vt (current_vt+1, 0);
	key_done = 1;
      }
      break;
    case SDLK_F1:
    case SDLK_F2:
    case SDLK_F3:
    case SDLK_F4:
    case SDLK_F5:
    case SDLK_F6:
    case SDLK_F7:
    case SDLK_F8:
    case SDLK_F9:
    case SDLK_F12:
      if (event->key.keysym.mod&magic_modifier) {
	int n = event->key.keysym.sym - SDLK_F1;
	select_vt (n, 0);
	key_done = 1;
      }
      break;
    default:
      shift_history = 0;
      ctrl_history = 0;
      break;
    }
    if (! key_done)  client_send_key (&event->key);
    break;

  case SDL_KEYUP:
    break;
    
  case SDL_QUIT:
    nchildren = 0;
    break;

  case SDL_USEREVENT:
    switch (event->user.code) {
    case EV_BLINK:		/* the blink timer expired */
      screen_toggle_cursor ();
      screen_toggle_blink ();
      break;
    case EV_FLASH:		/* the flash timer expired */
      screen_toggle_flash ();
      break;
    default:
      warning ("invalid user event code %d", event->user.code);
      break;
    }
    break;

  default:
    break;
  }
}

static void
listen_input (void)
{
  struct timeval  tv;
  fd_set  rfds;
  int  rc, i;
  
  FD_ZERO (&rfds);
  for (i=0; i<listeners_used; ++i)  FD_SET (listeners[i].fd, &rfds);

 retry_select:
  tv.tv_sec = 0;
  tv.tv_usec = 20000;
  rc = select (FD_SETSIZE, &rfds, NULL, NULL, &tv);
  if (rc < 0) {
    if (errno == EINTR)  goto retry_select;
    fatal ("select failed: %m");
  }

  for (i=0; i<listeners_used; ++i) {
    if (FD_ISSET (listeners[i].fd, &rfds)) {
      struct shell_client_data *scd = listeners[i].data;
      ssize_t  len;
      
      len = console_copy (scd->con, scd->fd);
      if (len < 0) {
	/* Child does no longer talk to us.  */
	if (errno != EIO)  error ("listen_input: console_copy failed (%m)");
      }
    }
  }
}

static void
main_loop (void)
{
  SDL_Event  event;

  while ((nchildren>0 || control_flag) && ! shutdown_flag) {
    int  status;
    pid_t  rc;
    
    listen_input ();
    
  retry_wait:
    rc = waitpid (WAIT_ANY, &status, WNOHANG);
    if (rc<0) {
      if (errno == EINTR)  goto retry_wait;
      error ("main_loop: waitpid failed: %m");
    } else if (rc>0) {
      int  i;

      for (i=0; i<8; ++i) {
	pid_t *p;
	p = client_apply (vt[i], shell_get_pid, NULL);
	if (*p == rc) {
	  message ("pid %d (vt%d) terminated", rc, i);
	  delete_client (vt[i]);
	  vt[i] = NULL;
	  for (current_vt=8-1; current_vt>=0; --current_vt) {
	    if (vt[current_vt])  break;
	  }
	  client_activate (dddlient);
	  break;
	}
      }
      if (i==8)  message ("pid %d (unknown) terminated", rc);
    }
    
    while (SDL_PollEvent (&event))  handle_event (&event);
    screen_update ();
  }
}

int
main (int argc, char **argv)
{
  int  error_flag = 0;
  int  hw_flag = 0;
  int  depth = 0;
  int  font_size = 15;
  const char *font_file = NULL;
  int  fullscreen_flag = 1;
  int  help_flag = 0;
  const char *log_file_name = NULL;
  int  metamode_flag = 0;
  int  verbose_flag = 0;
  int  version_flag = 0;
  const char *optarg;
  struct voption options [] = {
    { "bpp", 'b', NULL, 1, "BITS",
      "set the color depth" },
    { "ctype", 'c', NULL, 1, "STR",
      "set the child's LC_CTYPE to STR" },
    { "control", 'C', &control_flag, 0, NULL,
      "start with the control center activated" },
    { "exec", 'e', NULL, 1, "PROG",
      "start PROG instead of the default shell" },
    { "font", 'f', NULL, 1, "FILE",
      "use FILE as the font" },
    { "help", 'h', &help_flag, 0, NULL,
      "show this message" },
    { "logfile", 'l', NULL, 1, "LOG",
      "emit messages into the logfile LOG" },
    { "metamode", 'm', &metamode_flag, 0, NULL,
      "start the first shell in meta mode" },
    { "pt", 'p', NULL, 1, "SIZE",
      "set the font size in points" },
    { "hwsurface", 'q', &hw_flag, 0, NULL,
      "try to allocate a buffer in video memory" },
    { "verbose", 'v', &verbose_flag, 0, NULL,
      "emit messages to stdout" },
    { "version", 'V', &version_flag, 0, NULL,
      "show version information" },
    { "windowed", 'w', NULL, 0, NULL,
      "run in windowed mode (useful for debugging)" },
    { NULL, '\0', NULL, 0, NULL, NULL }
  };
  
  /****************************************
   * option parsing
   */
  open_options (argc, argv);
  do {
    int  c = options_get (options, &optarg, V_MIXED);
    if (c == -1)  break;

    switch (c) {
    case 'b':
      {
	char *tail;
	errno = 0;
	depth = strtol (optarg, &tail, 0);
	if (tail==optarg || *tail!=0 || errno || depth<1) {
	  error ("invalid color depth \"%s\"", optarg);
	  error_flag = 1;
	}
      }
      break;
    case 'c':
      child_LC_CTYPE = xstrdup (optarg);
      break;
    case 'e':
      default_shell = xstrdup (optarg);
      break;
    case 'f':
      font_file = xstrdup (optarg);
      break;
    case 'l':
      log_file_name = xstrdup (optarg);
      break;
    case 'p':
      {
	char *tail;
	errno = 0;
	font_size = strtol (optarg, &tail, 0);
	if (tail==optarg || *tail!=0 || errno || font_size<1) {
	  error ("invalid font size \"%s\"", optarg);
	  error_flag = 1;
	}
      }
      break;
    case 'w':
      fullscreen_flag = 0;
      break;
    case '\0':
      if (optarg)  error ("unknown option \"%s\"", optarg);
      error_flag = 1;
      break;
    default:
      fatal ("internal error while parsing options");
    }
  } while (! error_flag);
  
  if (version_flag) {
    puts ("jvterm " VERSION);
    puts ("Copyright (C) 2003, 2004 Jochen Voss <voss@seehuhn.de>");
    puts ("\
Jvterm comes with ABSOLUTELY NO WARRANTY, to the extent permitted by law.");
    puts ("\
You may redistribute copies of Jvterm under the terms of the GNU\n\
General Public License.  For more information about these matters, see\n\
the file named COPYING.");
    if (! error_flag)  exit (0);
  }

  if (error_flag || help_flag) {
    FILE *out = error_flag ? stderr : stdout;
    fprintf (out, "usage: %s [options]\n\n", argv[0]);
    options_show (options);
    fputs ("\nPlease report bugs to <voss@seehuhn.de>.\n", out);
    exit (error_flag);
  }
  close_options ();

  /****************************************
   * character encoding handling
   */
  parent_CHARSET = xstrdup (nl_langinfo (CODESET));
  if (child_LC_CTYPE) {
    setlocale (LC_CTYPE, child_LC_CTYPE);
    child_CHARSET = xstrdup (nl_langinfo (CODESET));
    setlocale (LC_CTYPE, "");
  } else {
    setlocale (LC_CTYPE, "");
    child_LC_CTYPE = xstrdup (setlocale (LC_CTYPE, NULL));
    child_CHARSET = xstrdup (nl_langinfo (CODESET));
  }
  
  /****************************************
   * initialise the mesage log
   */
  open_log ();
  open_std_logs (verbose_flag, log_file_name);
  message ("jvterm, version " VERSION);
  
  /****************************************
   * initialise the graphics engine
   */
  if ((SDL_Init (SDL_INIT_VIDEO|SDL_INIT_TIMER) == -1)) {
    fatal_sdl ("Could not initialize SDL");
  }
  atexit (quit_sdl);
  SDL_EnableKeyRepeat (SDL_DEFAULT_REPEAT_DELAY, SDL_DEFAULT_REPEAT_INTERVAL);
  SDL_EnableUNICODE (1);
  open_screen (depth, hw_flag, fullscreen_flag,
	       font_file, font_size);
  message ("caller's character encoding: %s", parent_CHARSET);
  message ("children's character encoding: %s", child_CHARSET);
  
  /****************************************
   * run the terminal emulator
   */
  nchildren = 0;
  {
    int i;
    for (i=0; i<8; ++i)  vt[i] = NULL;
  }
  dddlient = new_client (&control_client, NULL, 0);
  messages = new_client (&log_client, NULL, 1000);
  keyclient = new_client (&keyboard_client, NULL, 0);
  if (control_flag) {
    current_vt = 0;
    select_vt (8, 0);
  } else {
    select_vt (0, metamode_flag ? 2 : 1);
  }

  log_up_and_running ();

  ctrl_history = 0;
  shift_history = 0;
  
  blink_timer = SDL_AddTimer (500, blink_cb, NULL);
  main_loop ();

  /****************************************
   * clean up
   */
  {
    int  i;
    for (i=0; i<8; ++i) {
      if (vt[i])  delete_client (vt[i]);
    }
    delete_client (dddlient);
  }
  
  SDL_RemoveTimer (blink_timer);
  close_screen ();
  xfree (child_LC_CTYPE);
  /* 'quit_sdl' is run via 'at_exit'.  */

  close_std_logs ();
  close_log ();
  
  return  0;
}
