/* screen.c - low level screen access
 *
 * Copyright (C) 2003, 2005  Jochen Voss
 *
 * This program is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 2 of the License, or
 * (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 51 Franklin St, Fifth Floor, Boston, MA  02110-1301 USA
 *
 * $Id: screen.c 6725 2005-11-09 23:15:58Z voss $
 */

#ifdef HAVE_CONFIG_H
#  include <config.h>
#endif

#include <stdlib.h>
#include <string.h>
#include <assert.h>

#include <SDL.h>
#include <SDL_gfxPrimitives.h>

#include "base.h"
#include "backend.h"


const struct charcell  default_char = {
  c: ' ',
  attr: 0,
  colors: DEFAULT_COLOR
};


struct screen {
  SDL_Surface *buf;		/* the system's video buffer */
  struct font_info  fi;
  
  int  width;			/* screen width in pixels */
  int  height;			/* screen height in pixels */
  int  cols;			/* screen width in characters */
  int  rows;			/* screen height in characters */

  struct charcell *curstate;	/* the current state of the screen */
  struct charcell *nextstate; /* state of screen after next screen_update */
  int  image_dirty;
  int  margin_dirty;

  SDL_Surface *image;
  int  image_xoffs, image_yoffs;
  
  int  flash_state;		/* everything is current inverted */
  int  blink_state;		/* blinking text is currently inverted */
  int  cursor_state;		/* cursor is currently visible */
  int  curs_x, curs_y;		/* cursor position */

  int  sel_point;		/* the position of the selection point */
  int  sel_a;			/* the start of the selection */
  int  sel_b;			/* the end of the selection */
};
static  struct screen *screen;

/**********************************************************************
 * auxiliary functions
 */

static double
price (int n)
/* Get the approximate cost for issuing a call to SDL_UpdateRect,
 * which covers N glyph cells at once.  The numbers were determined
 * via measurements on my own system.  */
{
  const double timetable[15] = {
    0, 6.56604, 6.11321, 6.11321, 6.33962,
    6.56604, 6.79245, 7.24528, 8.15094, 9.28302,
    10.1887, 11.0943, 12, 13.1321, 14.2642
  };
  return  (n < 15) ? timetable[n] : n;
}

static int
different_p (struct charcell cur, struct charcell next)
/* Return true, iff the displayed character CUR differs currently from
 * the character NEXT.  This takes into account the current
 * flash_state, blink_state, and cursor_state.
 *
 * This must match the code in 'hw_copy_char'.  */
{
  if (cur.c != next.c || cur.colors != next.colors)  return 1;
  if (next.attr & 0x80) {
    if (screen->cursor_state && screen->sel_point < 0) {
      next.attr ^= JV_ATTR_REVERSE;
    }
    next.attr &= 0x7F;
  }
  if (next.attr&JV_ATTR_BLINK && screen->blink_state) {
    next.attr ^= JV_ATTR_REVERSE|JV_ATTR_BLINK;
  }
  if (screen->flash_state)  next.attr ^= JV_ATTR_REVERSE;
  return  cur.attr != next.attr;
}

static void
hw_copy_char (int i, int j)
/* Copy a character from the 'nextstate' table to the screen.
 * The 'curstate' table is updated to reflect the new screen state.
 *
 * This must match the code in 'different_p'.  */
{
  int  k = j*screen->cols + i;
  struct charcell  cc = screen->nextstate[k];
  SDL_Surface *glyph;
  int  rc;

  if (cc.attr & 0x80) {
    if (screen->cursor_state && screen->sel_point < 0) {
      cc.attr ^= JV_ATTR_REVERSE;
    }
    cc.attr &= 0x7F;
  }
  if (cc.attr&JV_ATTR_BLINK && screen->blink_state) {
    cc.attr ^= JV_ATTR_REVERSE|JV_ATTR_BLINK;
  }
  if (screen->flash_state)  cc.attr ^= JV_ATTR_REVERSE;

  glyph = font_get_glyph (&screen->fi, cc);
  if (glyph) {
    int  dx = screen->fi.cell_width;
    int  dy = screen->fi.cell_height;
    SDL_Rect  dest;

    dest.x = i*dx;
    dest.y = j*dy;
    dest.w = dx;
    dest.h = dy;
    rc = SDL_BlitSurface(glyph, NULL, screen->buf, &dest);
  } else {
    rc = -1;
  }

  if (rc == 0)  screen->curstate[k] = cc;
}

static void
hw_finish_char (int i, int j)
/* Ask libSDL to make our changes to character cell (i,j) visible.
 * This call is quite expensive.  */
{
  int  dx = screen->fi.cell_width;
  int  dy = screen->fi.cell_height;

  SDL_UpdateRect (screen->buf, i*dx, j*dy, dx, dy);
}

static void
hw_finish_row (int i0, int i1, int j)
/* Ask libSDL to make our changes to character cells (i0,j), (i0+1,j),
 * ..., (i1-1,j), visible.  This call is quite expensive.  */
{
  int  dx = screen->fi.cell_width;
  int  dy = screen->fi.cell_height;

  SDL_UpdateRect (screen->buf, i0*dx, j*dy, (i1-i0)*dx, dy);
}

static void
lock_cursor (void)
/* Remove the cursor marks in the character buffer.  This should be
 * called before either the cursor position or the character data at
 * the cursor position is changed.  */
{
  if (screen->curs_y >= 0) {
    int  k = screen->cols*screen->curs_y+screen->curs_x;
    screen->nextstate[k].attr &= 0x7F;
  }
}

static void
unlock_cursor (void)
/* Note the cursor position in the character buffer, again.  This
 * should be called after either the cursor position or the character
 * data at the cursor position is changed.  */
{
  if (screen->curs_y >= 0) {
    int  k = screen->cols*screen->curs_y+screen->curs_x;
    screen->nextstate[k].attr |= 0x80;
  }
}

static void
screen_mark_cell_dirty (int i)
{
  screen->curstate[i].c = 0;
  screen->curstate[i].attr = 0;
  screen->curstate[i].colors = 0;
}

static void
screen_mark_dirty (void)
{
  int  i;

  for (i=0; i<screen->cols * screen->rows; ++i) {
    screen_mark_cell_dirty (i);
  }
  screen->image_dirty = 1;
  screen->margin_dirty = 1;
}

/**********************************************************************
 * selection handling
 */

static int
selection_isinside (int i, int j)
{
  int  pos;
  
  pos = j*screen->cols + i;
  return  (pos >= screen->sel_a && pos < screen->sel_b);
}

static int
selection_istop (int i, int j)
{
  int  pos;
  
  if (! selection_isinside(i,j))  return 0;
  pos = j*screen->cols + i;
  return  (pos < screen->sel_a+screen->cols);
}

static int
selection_isbottom (int i, int j)
{
  int  pos;
  
  if (! selection_isinside(i,j))  return 0;
  pos = j*screen->cols + i;
  return  (pos > (screen->sel_b-1)-screen->cols);
}

static int
selection_isleft (int i, int j)
{
  int  pos;
  
  if (! selection_isinside(i,j))  return 0;
  pos = j*screen->cols + i;
  return  (pos == screen->sel_a || i == 0);
}

static int
selection_isright (int i, int j)
{
  int  pos;
  
  if (! selection_isinside(i,j))  return 0;
  pos = j*screen->cols + i;
  return  (pos == screen->sel_b-1 || i == screen->cols-1);
}

/**********************************************************************
 * public screen functions
 */

void
open_screen (int depth, int hw_flag, int fullscreen_flag,
	     const char *font_file, double font_size)
/* Initialise the SDL screen access.
 * DEPTH is the color depth (or 0 for driver default).
 * A non-zero HW_FLAG means to allocate a SDL_HWSURFACE.
 * FULLSCREEN_FLAG switches to fullscreen mode instead of creating a window.
 * FONT_SIZE is the font size to use in points.  */
{
  SDL_Rect **modes;
  int  width, height;
  Uint32  flags;
  int  i, n;

  screen = xnew (struct screen);

  /* load the font early, so we can emit potential errors
   * before SDL is initialised.  */
  load_font (font_file, font_size, &screen->fi);

  flags = SDL_HWPALETTE;
  if (depth == 0) {
    flags |= SDL_ANYFORMAT;
    depth = 16;
  }
  if (hw_flag)  flags |= SDL_HWSURFACE;
  if (fullscreen_flag) {
    flags |= SDL_FULLSCREEN;
  }

  modes = SDL_ListModes (NULL, flags);
  if (modes == (SDL_Rect **)0) {
    /* No modes available?  Let's try something safe.  */
    warning ("SDL reports no available video modes");
    if (fullscreen_flag) {
      width = 1024;
      height = 768;
    } else {
      width = 800;
      height = 600;
    }
  } else if (modes == (SDL_Rect **)-1 ) {
    /* arbitrary modes available */
    if (fullscreen_flag) {
      width = 100 * screen->fi.cell_width;
      height = 37 * screen->fi.cell_height;
    } else {
      width = 80 * screen->fi.cell_width;
      height = 25 * screen->fi.cell_height;
    }
  } else{
    i = 0;
    while (modes[i+1]
	   && (modes[i]->w > 200*screen->fi.cell_width
	       || modes[i]->h > 80*screen->fi.cell_height)) ++i;
    width = modes[i]->w;
    height = modes[i]->h;
  }
  
  screen->buf = SDL_SetVideoMode (width, height, depth, flags);
  if (! screen->buf)  fatal_sdl ("could not set video mode");
  if (fullscreen_flag)  SDL_ShowCursor (SDL_DISABLE);
  
  screen->width = width;
  screen->height = height;
  screen->cols = width / screen->fi.cell_width;
  screen->rows = height / screen->fi.cell_height;
  message ("screen dimensions: %dx%d pixels, %dx%d characters",
	   screen->width, screen->height, screen->cols, screen->rows);

  n = screen->cols * screen->rows;
  screen->curstate = xmalloc (n * sizeof(struct charcell));
  screen->nextstate = xmalloc (n * sizeof(struct charcell));
  for (i=0; i<n; ++i) {
    screen_mark_cell_dirty (i);
    screen->nextstate[i] = default_char;
  }
  screen->image_dirty = 1;
  screen->margin_dirty = 1;

  screen->image = NULL;
  screen->image_xoffs = 0;
  screen->image_yoffs = 0;
  
  screen->flash_state = 0;
  screen->blink_state = 0;
  screen->cursor_state = 0;
  screen->curs_x = 0;
  screen->curs_y = -1;

  screen->sel_point = -1;
  screen->sel_a = screen->sel_b = 0;

  SDL_WM_SetCaption ("JVTerm terminal emulator", "JVTerm");
}

void
close_screen (void)
/* Unload the fonts and free the SDL screen surface.  */
{
  xfree (screen->nextstate);
  xfree (screen->curstate);
  unload_font ();
  SDL_FreeSurface (screen->buf);
  xfree (screen);
  screen = NULL;
}

int
screen_width (void)
{
  return  screen->width;
}

int
screen_height (void)
{
  return  screen->height;
}

int
screen_get_cols (void)
{
  return  screen->cols;
}

int
screen_get_rows (void)
{
  return  screen->rows;
}

void
screen_update_image (void)
/* screen_update function while in image display mode */
{
  SDL_Surface *im;
  SDL_Rect  src, dst;

  if (! screen->image_dirty)  return;
  
  im = screen->image;

  if (im->w < screen->width || im->h < screen->height) {
    SDL_FillRect (screen->buf, NULL,
		  SDL_MapRGB (screen->buf->format, 0, 0, 0));
  }

  if (im->w < screen->width) {
    src.x = 0;
    src.w = im->w;
    dst.x = 0.5*(screen->width-src.w);
  } else {
    src.x = screen->image_xoffs;
    src.w = screen->width;
    dst.x = 0;
  }

  if (im->h < screen->height) {
    src.y = 0;
    src.h = im->h;
    dst.y = 0.5*(screen->height-src.h);
  } else {
    src.y = screen->image_yoffs;
    src.h = screen->height;
    dst.y = 0;
  }

  SDL_BlitSurface (im, &src, screen->buf, &dst);
  SDL_UpdateRect (screen->buf, 0, 0, 0, 0);

  screen->image_dirty = 0;
}

void
screen_update_text (void)
/* screen_update function while in text mode */
{
  int  rows = screen->rows;
  int  cols = screen->cols;
  int  i, j;

  for (j=0; j<rows; ++j) {
    char  updated[cols];
    int  min, max;
    int  count = 0;

    for (i=0; i<cols; ++i) {
      int  k = j*cols + i;
      if (different_p (screen->curstate[k], screen->nextstate[k])) {
	int  dx = screen->fi.cell_width;
	int  dy = screen->fi.cell_height;
	
	updated[count++] = i;
	hw_copy_char (i, j);

	if (screen->cursor_state && i+j*screen->cols == screen->sel_point) {
	  boxRGBA (screen->buf, i*dx, j*dy, (i+0.4)*dx, (j+1)*dy-1,
		   196, 32, 32, 255);
	}
	if (screen->cursor_state
	    && i+j*screen->cols == screen->sel_point - 1
	    && screen->sel_point == screen->cols*screen->rows) {
	  boxRGBA (screen->buf, (i+0.6)*dx, j*dy, (i+1)*dx-1, (j+1)*dy-1,
		   196, 32, 32, 255);
	}
	if (selection_isinside (i,j)) {
	  boxRGBA (screen->buf, i*dx, j*dy, (i+1)*dx-1, (j+1)*dy-1,
		   196, 32, 32, 100);
	}
	if (selection_istop (i,j)) {
	  hlineRGBA (screen->buf, i*dx, (i+1)*dx-1, j*dy,
		     255, 64, 32, 255);
	}
	if (selection_isbottom (i,j)) {
	  hlineRGBA (screen->buf, i*dx, (i+1)*dx-1, (j+1)*dy-1,
		     255, 64, 32, 255);
	}
	if (selection_isleft (i,j)) {
	  vlineRGBA (screen->buf, i*dx, j*dy, (j+1)*dy-1,
		     255, 64, 32, 255);
	}
	if (selection_isright (i,j)) {
	  vlineRGBA (screen->buf, (i+1)*dx-1, j*dy, (j+1)*dy-1,
		     255, 64, 32, 255);
	}
      }
    }
    if (count == 0)  continue;
    
    min = updated[0];
    max = updated[count-1]+1;
    if (price(1)*count < price(max-min)) {
      for (i=0; i<count; ++i)  hw_finish_char (updated[i], j);
    } else {
      hw_finish_row (min, max, j);
    }
  }
}

void
screen_update (void)
/* Make sure that the current screen display matches the buffer
 * contents.  This must be called periodically to make the new screen
 * contents visible.  */
{
  if (screen->image) {
    screen_update_image ();
  } else {
    screen_update_text ();
  }
}

void
screen_set_cursor (int x, int y)
/* Set the cursor to a new position.
 * To disable the cursor move it vertically outside the screen.  */
{
  lock_cursor ();
  screen->curs_x = x;
  if (y >= screen->rows)  y = -1;
  screen->curs_y = y;
  unlock_cursor ();

  if (y >= 0) {
    screen->cursor_state = 1;
  } else {
    screen->cursor_state = 0;
  }
}

void
screen_toggle_cursor (void)
/* Toggle the visibility for a blinking cursor.  */
{
  if (screen->curs_y >= 0) {
    screen->cursor_state = ! screen->cursor_state;
  }
  if (screen->sel_point >= 0) {
    int  n = screen->cols * screen->rows;
    if (screen->sel_point < n) {
      screen_mark_cell_dirty (screen->sel_point);
    } else {
      screen_mark_cell_dirty (n-1);
    }
  }
  screen_update ();
}

void
screen_toggle_blink (void)
/* Toggle display of blinking text.
 * This should be called periodically.  */
{
  screen->blink_state = ! screen->blink_state;
  screen_update ();
}

void
screen_toggle_flash (void)
/* Invert the screen contents.
 * This is meant to implement a visible bell.
 * It should be called an even number of times.  */
{
  screen->flash_state = ! screen->flash_state;
  screen_update ();
}

void
screen_clear (void)
/* Fill all character cells with the default character.
 * This function does not influence the cursor.  */
{
  int  i, n;

  lock_cursor ();
  n = screen->cols * screen->rows;
  for (i=0; i<n; ++i)  screen->nextstate[i] = default_char;
  unlock_cursor ();
}

struct charcell
screen_get_cell (int i, int j)
{
  int  k = j*screen->cols + i;
  return   screen->curstate[k];
}

void
screen_draw_cell (struct charcell cc, int x, int y)
/* Write CC into the character cell at position (X,Y).
 * This function does not influence the cursor.  */
{
  int  cflag = (x == screen->curs_x && y == screen->curs_y);

  if (cflag)  lock_cursor ();
  screen->nextstate[screen->cols*y+x] = cc;
  if (cflag)  unlock_cursor ();
}

void
screen_scroll_up (int y0, int y1)
/* Move the screen lines y0+1, y0+2, ..., y1-1 up, overwriting line y0
 * and leaving line y1-1 empty.  */
{
  int  cols = screen->cols;
  int  base;
  int  i;

  lock_cursor ();
  
  base = y0*cols;
  memmove (screen->nextstate+base, screen->nextstate+(base+cols),
	   (y1-y0-1)*cols*sizeof(struct charcell));
  base = (y1-1)*cols;
  for (i=0; i<cols; ++i)  screen->nextstate[base+i] = default_char;

  unlock_cursor ();
}

void
screen_scroll_down (int y0, int y1)
/* Move screen lines y0, y0+1, ..., y1-2 down, overwriting line y1-1
 * and leaving line y0 empty.  */
{
  int  cols = screen->cols;
  int  base;
  int  i;

  lock_cursor ();

  base = y0*cols;
  memmove (screen->nextstate+(base+cols), screen->nextstate+base,
	   (y1-y0-1)*cols*sizeof(struct charcell));
  for (i=0; i<cols; ++i)  screen->nextstate[base+i] = default_char;

  unlock_cursor ();
}

void
screen_set_selection (int point_x, int point_y, int mark_x, int mark_y)
/* Set point and mark for the selection.  To disable the selection
 * call this function with POINT_Y<0.  */
{
  int  cols = screen->cols;
  int  a, b, i, n;

  n = screen->cols*screen->rows;
  a = point_x + cols*point_y;
  b = mark_x + cols*mark_y;
  if (a>b) {
    int  tmp = a;
    a = b;
    b = tmp;
  }

  if (screen->sel_point >= 0) {
    if (screen->sel_point < screen->cols*screen->rows) {
      screen_mark_cell_dirty (screen->sel_point);
    } else {
      screen_mark_cell_dirty (screen->sel_point-1);
    }
  }

  if (point_y < 0) {
    for (i=max(screen->sel_a, 0); i<min(screen->sel_b,n); ++i)
      screen_mark_cell_dirty (i);
    screen->sel_point = -1;
    screen->sel_a = screen->sel_b = 0;
    return;
  }

  for (i=max(min(a,screen->sel_a),0); i<min(max(b,screen->sel_b),n); ++i) {
    screen_mark_cell_dirty (i);
  }
  screen->sel_point = point_x + cols*point_y;
  if (screen->sel_point < screen->cols*screen->rows) {
    screen_mark_cell_dirty (screen->sel_point);
  } else {
    screen_mark_cell_dirty (screen->sel_point-1);
  }
  screen->sel_a = a;
  screen->sel_b = b;

  screen->cursor_state = 1;
}

/**********************************************************************
 * image handling
 */

void
screen_set_image (SDL_Surface *image, int xoffs, int yoffs)
{
  screen->image = image;
  screen->image_xoffs = xoffs;
  screen->image_yoffs = yoffs;
  if (image) {
    screen->image_dirty = 1;
  } else {
    screen_mark_dirty ();
  }
}

/**********************************************************************
 * unit testing
 */

#ifdef TEST_SCREEN

#include <stdio.h>
#include <unistd.h>
#include <jvrand.h>

int
main (int argc, char **argv)
{
  int  error_flag = 0;
  int  hw_flag = 0;
  int  depth = 0;
  int  font_size = 15;
  const char *font_file = NULL;
  int  fullscreen_flag = 1;

  struct charcell  cc;
  Uint32  t0, t1;
  int  i, j, l;
  
  do {
    int  c;
    c = getopt (argc, argv, "db:f:p:qw");
    if (c == -1)  break;

    switch (c) {
    case 'b':
      depth = strtol (optarg, NULL, 0);
      break;
    case 'f':
      font_file = xstrdup (optarg);
      break;
    case 'q':
      hw_flag = 1;
      break;
    case 'p':
      font_size = strtol (optarg, NULL, 0);
      break;
    case 'w':
      fullscreen_flag = 0;
      break;
    default:
      error_flag = 1;
    }
  } while (! error_flag);
  
  if ((SDL_Init (SDL_INIT_VIDEO|SDL_INIT_TIMER) == -1)) {
    fatal_sdl ("Could not initialize SDL");
  }
  
  open_screen (depth, hw_flag, fullscreen_flag, font_file, font_size);
  for (i=0; i<screen->cols; ++i) {
    for (j=0; j<screen->rows; ++j) {
      cc.c = jvrand_discrete ('A', 'Z'+1);
      cc.attr = 0;
      cc.colors = DEFAULT_COLOR;
      screen_draw_cell (cc, i, j);
    }
  }
  screen_update ();

  puts ("test 1: drawing the same cell 10000 times");
  for (l=1; l<=5; ++l) {
    t0 = SDL_GetTicks ();
    for (i=0; ((t1 = SDL_GetTicks ())<t0+2000) && i<10000; ++i)
      hw_finish_char (7, 2);
    printf ("%dms -> %.1f cps\n", t1-t0, i*1000.0/(t1-t0));
  }

  puts ("test 2: drawing the same line 1000 times");
  for (l=1; l<=5; ++l) {
    t0 = SDL_GetTicks ();
    for (i=0; ((t1 = SDL_GetTicks ())<t0+2000) && i<1000; ++i)
      hw_finish_row (0, screen->cols, 1);
    printf ("%dms -> %.1f cps\n", t1-t0, screen->cols*i*1000.0/(t1-t0));
  }

  puts ("test 3: highlighting random selections");
  for (l=1; l<=5; ++l) {
    t0 = SDL_GetTicks ();
    for (i=0; ((t1 = SDL_GetTicks ())<t0+2000) && i<100; ++i) {
      screen_set_selection (jvrand_discrete(0,screen->cols),
			    jvrand_discrete(0,screen->rows),
			    jvrand_discrete(0,screen->cols),
			    jvrand_discrete(0,screen->rows));
      screen_update ();
    }
    printf ("%dms -> %.1f sps\n", t1-t0, i*1000.0/(t1-t0));
  }
  screen_set_selection (0, -1, 0, 0);

  puts ("test 4: scrolling almost empty lines");
  screen_scroll_up (0, screen->rows);
  cc.c = '*';
  cc.attr = 0;
  cc.colors = DEFAULT_COLOR;
  t0 = SDL_GetTicks ();
  for (i=0; ((t1 = SDL_GetTicks ())<t0+2000) && i<1000; ++i) {
    screen_draw_cell (cc, jvrand_discrete(0, screen->cols),
		      screen->rows-1);
    screen_scroll_up (0, screen->rows);
    screen_update ();
  }
  printf ("%dms -> %.1f lps\n", t1-t0, i*1000.0/(t1-t0));

  close_screen ();

  SDL_Quit();

  return  0;
}

#endif
