/* terminal.c - create a pseudo terminal for the child process
 *
 * Copyright (C) 2003, 2005  Jochen Voss
 *
 * This program is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 2 of the License, or
 * (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 51 Franklin St, Fifth Floor, Boston, MA  02110-1301 USA
 *
 * $Id: terminal.c 6725 2005-11-09 23:15:58Z voss $
 */

#ifdef HAVE_CONFIG_H
#  include <config.h>
#endif

#include <stdlib.h>
#include <stdio.h>
#include <sys/types.h>
#include <unistd.h>
#include <fcntl.h>
#include <sys/ioctl.h>
#include <signal.h>
#include <locale.h>
#include <pwd.h>
#include <errno.h>

#include "base.h"
#include "frontend.h"


static int
open_pty_pair (int *amaster, int *aslave)
/* Allocate a new pty pair.  Return 0 on failure and 1 on success.  */
{
#if HAVE_GETPT
  int  master, slave;
  char *name;

  master = getpt ();
  if (master < 0)  return 0;
     
  if (grantpt (master) < 0 || unlockpt (master) < 0)  goto close_master;
  name = ptsname (master);
  if (name == NULL)  goto close_master;
     
  slave = open (name, O_RDWR);
  if (slave == -1)  goto close_master;
  
  *amaster = master;
  *aslave = slave;
  return 1;
     
 close_slave:
  close (slave);
     
 close_master:
  close (master);
  return 0;
#else
  int  rc;

  rc = openpty (amaster, aslave, NULL, NULL, NULL);
  return  (rc != -1);
#endif
}

static void
set_window_size (int fd, int width, int height)
/* stolen from iterm-0.5 */
{
  struct winsize  wsize;

  wsize.ws_col = width;
  wsize.ws_row = height;
  wsize.ws_xpixel = 0;		/* TODO */
  wsize.ws_ypixel = 0;		/* TODO */
  
  if(ioctl(fd,TIOCSWINSZ,&wsize)) {
    perror ("TIOCSWINSZ");
    exit (1);
  }
}

static void
setup_child (int  fd, struct shell_info *info)
/* Replace the current process with a shell.
 * This is only run in the child process.  */
{
  char  user_buffer [32];
  char  logname_buffer [32];
  char  home_buffer [256];
  char  shell_buffer [128];
  char  columns_buffer [32];
  char  lines_buffer [32];
  char  lc_ctype_buffer [256];
  char *env[16];
  int  env_count;
  const char *shell = NULL;
  uid_t  uid;
  struct passwd  pw, *pw_ptr;
  char buffer[512];

  env_count = 0;
  uid = getuid ();
  getpwuid_r (uid, &pw, buffer, 512, &pw_ptr);

  if (pw_ptr) {
    snprintf (user_buffer, 32, "USER=%s", pw_ptr->pw_name);
    env[env_count++] = user_buffer;
    snprintf (logname_buffer, 32, "LOGNAME=%s", pw_ptr->pw_name);
    env[env_count++] = logname_buffer;
    if (pw_ptr->pw_dir) {
      snprintf (home_buffer, 256, "HOME=%s", pw_ptr->pw_dir);
      env[env_count++] = home_buffer;
    }
    if (pw_ptr->pw_shell)  shell = pw_ptr->pw_shell;
  }
  if (! shell)  shell = "/bin/sh";
  if (info->shell)  shell = info->shell;
  snprintf (shell_buffer, 128, "SHELL=%s", shell);
  env[env_count++] = shell_buffer;
  snprintf (columns_buffer, 32, "COLUMNS=%d", info->cols);
  env[env_count++] = columns_buffer;
  snprintf (lines_buffer, 32, "LINES=%d", info->rows);
  env[env_count++] = lines_buffer;
  snprintf (lc_ctype_buffer, 256, "LC_CTYPE=%s", info->lc_ctype);
  env[env_count++] = lc_ctype_buffer;

  env[env_count++] = "PATH=/usr/local/bin:/usr/bin:/bin";
  env[env_count++] = info->meta_mode ? "TERM=jvterm3-km" : "TERM=jvterm3";
  env[env_count++] = NULL;
  
  signal (SIGHUP, SIG_DFL);
  signal (SIGINT, SIG_DFL);
  signal (SIGQUIT, SIG_DFL);
  signal (SIGTERM, SIG_DFL);
  signal (SIGTTIN, SIG_DFL);
  signal (SIGTTOU, SIG_DFL);
  signal (SIGPIPE, SIG_DFL);
  signal (SIGCHLD, SIG_DFL);

  /* TODO: we should check for errors in the following code.
   * But how do we communicate them back?
   * idea: we could use a special protocol on the terminal,
   * which is disabled via some code before the shell starts.  */
  
  /* detach from the current controlling tty */
  setsid ();

  /* make the pty the new controlling tty */
  ioctl (fd, TIOCSCTTY, 0);
  
  set_window_size (fd, info->cols, info->rows);
  
  dup2 (fd, STDIN_FILENO);
  dup2 (fd, STDOUT_FILENO);
  dup2 (fd, STDERR_FILENO);
  close (fd);

  {
    char *args[] = {
      "-",
      NULL
    };
    execve (shell, args, env);
  }
  perror ("execl child");	/* TODO: this error message will not be seen */
  _exit (1);
}

int
spawn_shell (struct shell_info *info, pid_t *pid_return)
/* Create a new process running a shell.
 * Return the file descriptor for our end of the controlling tty.  */
{
  pid_t  child;
  int  my_fd, his_fd;

  if (! open_pty_pair(&my_fd, &his_fd)) {
    fatal ("cannot open pty pair");
  }
  
  fflush (stdout);
  fflush (stderr);

  child = fork ();
  if (child == -1) {
    perror ("fork");
    exit (1);
  } else if (child == 0) {
    close (my_fd);
    setup_child (his_fd, info);
    /* does not return */
  }
  close (his_fd);

  if (pid_return) *pid_return = child;
  return  my_fd;
}
